import math
import warnings
from operator import xor
from typing import Literal, Optional, Union, overload

import awkward as ak
import numba as nb
import numpy as np
import vector
import vector.backends.awkward as vec_ak

vector.register_awkward()

from .._utils import _extract_index, _flat_to_numpy, _recover_shape
from ..typing import FloatLike, IntLike

TypeObjPosition = Union[vector.VectorObject3D, tuple[float, float, float]]
TypeObjMomentum = Union[vector.MomentumObject3D, tuple[float, float, float]]
TypeAwkPosition = Union[ak.Array, vector.VectorObject3D, tuple[float, float, float]]


def _regularize_obj_position(
    args: Union[TypeObjPosition, tuple[TypeObjPosition]],
) -> vector.VectorObject3D:
    """
    Regularizes the position argument to always return a VectorObject3D.

    Args:
        args: A tuple of (x, y, z), a VectorObject3D, an ak.Record, or a tuple
            containing one of them.

    Returns:
        A VectorObject3D representing the position point.
    """
    if isinstance(args, tuple) and len(args) == 1:
        args = args[0]

    if isinstance(args, vector.VectorObject3D):
        return args

    if isinstance(args, ak.Record):
        return vector.VectorObject3D(x=args["x"], y=args["y"], z=args["z"])

    # tuple
    point = args if args is not None else (0, 0, 0)

    assert len(point) == 3, "Pivot must be a tuple of (x, y, z)."
    return vector.VectorObject3D(x=point[0], y=point[1], z=point[2])


def _regularize_obj_momentum(
    args: Union[TypeObjMomentum, tuple[TypeObjMomentum]],
) -> vector.MomentumObject3D:
    """
    Regularizes the momentum argument to always return a MomentumObject3D.

    Args:
        args: A tuple of (px, py, pz), a MomentumObject3D, an ak.Record, or a tuple
            containing one of them.

    Returns:
        A MomentumObject3D representing the momentum vector.
    """
    if isinstance(args, tuple) and len(args) == 1:
        args = args[0]

    if isinstance(args, vector.MomentumObject3D):
        return args

    if isinstance(args, ak.Record):
        return vector.MomentumObject3D(
            px=args["px"],
            py=args["py"],
            pz=args["pz"],
        )

    # tuple
    momentum = args
    assert len(momentum) == 3, "Momentum must be a tuple of (px, py, pz)."
    return vector.MomentumObject3D(px=momentum[0], py=momentum[1], pz=momentum[2])


def _change_pivot(
    r,
    old_dr,
    old_phi0,
    old_dz,
    kappa,
    tanl,
    old_error,
    old_pivot,
    new_pivot,
):
    """
    Change the pivot point of the helix and transform its parameters accordingly.
    The transformation is based on the BOSS `Helix::pivot` method.

    This method will be called by the `change_pivot` method of `HelixObject`, `HelixAwkwardRecord` and `HelixAwkwardArray`.

    Args:
        r: Circular radius of the helix.
        old_dr: Current dr of the helix.
        old_phi0: Current phi0 of the helix.
        old_dz: Current dz of the helix.
        kappa: Current kappa of the helix.
        tanl: Current tanl of the helix.
        old_error: Current error matrix of the helix, if available.
        old_pivot: Current pivot point of the helix, either as a VectorObject3D or a tuple of (x, y, z).
        new_pivot: New pivot point of the helix, either as a VectorObject3D or a tuple of (x, y, z).

    Returns:
        A tuple containing the new dr, new phi0, new dz, and the transformed error matrix (if available).
    """
    if isinstance(old_pivot, vector.VectorObject3D):
        old_dist = vector.obj(rho=old_dr + r, phi=old_phi0)
    elif isinstance(old_pivot, vector.VectorNumpy3D):
        old_dist = vector.arr({"rho": old_dr + r, "phi": old_phi0})
    else:
        raise TypeError(f"Unsupported old_pivot type: {type(old_pivot)}. ")

    center = old_pivot.to_2D() + old_dist

    new_dist: vector.Vector2D = center - new_pivot.to_2D()

    new_dr = new_dist.rho - r
    new_phi0 = new_dist.phi % (2 * np.pi)

    if isinstance(new_phi0, np.ndarray):
        dphi = np.unwrap(new_phi0 - old_phi0)
    else:
        dphi = (new_phi0 - old_phi0) % (2 * np.pi)
        if dphi > np.pi:
            dphi -= 2 * np.pi
    new_dz = old_pivot.z + old_dz - r * tanl * dphi - new_pivot.z

    # transform error matrix
    if old_error is not None:
        cos_dphi = np.cos(dphi)
        sin_dphi = np.sin(dphi)

        rdr = r + old_dr
        rdrpr = 1 / (r + new_dr)

        jacobian: np.ndarray = np.zeros_like(old_error)  # (tracks, i, j)
        if jacobian.ndim == 3:
            jacobian = jacobian.transpose(1, 2, 0)  # (i, j, tracks)

        jacobian[0, 0] = cos_dphi
        jacobian[0, 1] = rdr * sin_dphi
        jacobian[0, 2] = r / kappa * (1 - cos_dphi)

        jacobian[1, 0] = -rdrpr * sin_dphi
        jacobian[1, 1] = rdr * rdrpr * cos_dphi
        jacobian[1, 2] = r / kappa * rdrpr * sin_dphi

        jacobian[2, 2] = 1

        jacobian[3, 0] = r * rdrpr * tanl * sin_dphi
        jacobian[3, 1] = r * tanl * (1 - rdr * rdrpr * cos_dphi)
        jacobian[3, 2] = r / kappa * tanl * (dphi - r * rdrpr * sin_dphi)
        jacobian[3, 3] = 1
        jacobian[3, 4] = -r * dphi

        jacobian[4, 4] = 1

        if jacobian.ndim == 3:
            jacobian = jacobian.transpose(2, 0, 1)  # (tracks, i, j)
            new_error = jacobian @ old_error @ jacobian.transpose(0, 2, 1)
        else:
            new_error = jacobian @ old_error @ jacobian.T
    else:
        new_error = None

    return new_dr, new_phi0, new_dz, new_error


def _obj_isclose(self, other, *, rtol: float, atol: float, equal_nan: bool) -> bool:
    kwargs = {"rtol": rtol, "atol": atol, "equal_nan": equal_nan}
    other = other.change_pivot(self.pivot)

    # when `self` is an ak.Record, its pivot needs to be converted to a VectorObject3D
    self_pivot = vector.obj(x=self.pivot.x, y=self.pivot.y, z=self.pivot.z)
    other_pivot = vector.obj(x=other.pivot.x, y=other.pivot.y, z=other.pivot.z)

    condition = (
        np.isclose(self.dr, other.dr, **kwargs)
        and np.isclose(self.phi0, other.phi0, **kwargs)
        and np.isclose(self.kappa, other.kappa, **kwargs)
        and np.isclose(self.tanl, other.tanl, **kwargs)
        and np.isclose(self.dz, other.dz, **kwargs)
        and np.isclose(np.abs(self_pivot - other_pivot), 0, **kwargs)
    )

    if self.error is not None and other.error is not None:
        condition = condition and np.allclose(self.error, other.error, **kwargs)

    return bool(condition)


def _arr_isclose(self, other, *, rtol: float, atol: float, equal_nan: bool) -> ak.Array:
    kwargs = {"rtol": rtol, "atol": atol, "equal_nan": equal_nan}
    other = other.change_pivot(self.pivot)

    condition = ak.ones_like(self.dr, dtype=bool)
    for f in [
        "dr",
        "phi0",
        "kappa",
        "tanl",
        "dz",
    ]:
        condition = condition & ak.isclose(self[f], other[f], **kwargs)

    # Check pivot separately
    if isinstance(self, ak.Record):
        self_pivot = ak.Record(
            {"x": self.pivot.x, "y": self.pivot.y, "z": self.pivot.z},
            with_name="Vector3D",
        )
        other_pivot = ak.Record(
            {"x": other.pivot.x, "y": other.pivot.y, "z": other.pivot.z},
            with_name="Vector3D",
        )
    else:
        self_pivot = self.pivot
        other_pivot = other.pivot

    condition = condition & ak.isclose(np.abs(self_pivot - other_pivot), 0, **kwargs)

    if "error" in self.fields and "error" in other.fields:
        condition = condition & ak.all(
            ak.all(ak.isclose(self.error, other.error, **kwargs), axis=-1), axis=-1
        )

    return condition


###############################################################################################


class HelixObject:
    def __init__(
        self,
        dr: float,
        phi0: float,
        kappa: float,
        dz: float,
        tanl: float,
        *,
        error: np.ndarray = None,
        pivot: TypeObjPosition = (0, 0, 0),
    ):
        self.dr = float(dr)
        self.phi0 = float(phi0)
        self.kappa = float(kappa)
        self.dz = float(dz)
        self.tanl = float(tanl)
        self.error = error
        self.pivot = _regularize_obj_position(pivot)

    @property
    def radius(self) -> float:
        """
        Circular radius of the helix (in cm).
        """
        return 1000 / 2.99792458 / np.abs(self.kappa)

    @property
    def momentum(self) -> vector.MomentumObject3D:
        """
        Momentum of the helix as a 3D vector.
        Note that the momentum is relative to the pivot, so once the pivot is changed,
        the momentum will also change accordingly.
        """
        pt = 1 / abs(self.kappa)
        pz = pt * self.tanl
        phi = (self.phi0 + np.pi / 2) % (2 * np.pi)
        return vector.obj(pt=pt, phi=phi, pz=pz)

    @property
    def position(self) -> vector.VectorObject3D:
        return vector.VectorObject3D(
            x=self.dr * math.cos(self.phi0),
            y=self.dr * math.sin(self.phi0),
            z=self.dz,
        )

    @property
    def charge(self) -> int:
        """
        Returns the charge of the helix.
        """
        return 1 if self.kappa > 1e-10 else -1 if self.kappa < -1e-10 else 0

    def change_pivot(self, *args):
        """
        Change the pivot point of the helix.
        The transformation refers to `Helix` class in `BOSS`.
        """
        # transform helix parameters
        r = self.radius

        old_dr = self.dr
        old_phi0 = self.phi0
        old_dz = self.dz
        tanl = self.tanl
        kappa = self.kappa

        old_pivot = self.pivot
        new_pivot = _regularize_obj_position(args)

        new_dr, new_phi0, new_dz, new_error = _change_pivot(
            r=r,
            old_dr=old_dr,
            old_phi0=old_phi0,
            old_dz=old_dz,
            kappa=kappa,
            tanl=tanl,
            old_error=self.error,
            old_pivot=old_pivot,
            new_pivot=new_pivot,
        )

        return HelixObject(
            dr=new_dr,
            phi0=new_phi0,
            kappa=kappa,
            dz=new_dz,
            tanl=tanl,
            error=new_error,
            pivot=new_pivot,
        )

    def __repr__(self) -> str:
        return f"Bes3Helix(dr={self.dr:.3f}, phi0={self.phi0:.3f}, kappa={self.kappa:.3f}, tanl={self.tanl:.3f}, dz={self.dz:.3f})"

    def isclose(
        self,
        other: "HelixObject",
        *,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        equal_nan: bool = False,
    ) -> bool:
        if xor(
            (self.error is not None),
            (other.error is not None),
        ):
            warnings.warn(
                "One of the helix records has an error matrix, but the other does not. "
                "Ignoring error matrix for isclose check.",
                UserWarning,
            )

        return _obj_isclose(self, other, rtol=rtol, atol=atol, equal_nan=equal_nan)


###############################################################################################


def _check_kwargs_used_up(left_kwargs):
    if len(left_kwargs):
        warnings.warn(
            "Ignoring additional keyword arguments: " + ", ".join(left_kwargs.keys()),
            UserWarning,
        )


# case 1
@overload
def helix_obj(
    dr: float,
    phi0: float,
    kappa: float,
    dz: float,
    tanl: float,
    *,
    error: Optional[np.ndarray] = None,
    pivot: TypeObjPosition = (0, 0, 0),
) -> HelixObject: ...


# case 2
@overload
def helix_obj(
    *,
    params: tuple[float, float, float, float, float],
    error: Optional[np.ndarray] = None,
    pivot: TypeObjPosition = (0, 0, 0),
) -> HelixObject: ...


# case 3
@overload
def helix_obj(
    *,
    momentum: TypeObjMomentum,
    position: TypeObjPosition,
    charge: Literal[-1, 1],
    error: Optional[np.ndarray] = None,
    pivot: TypeObjPosition = (0, 0, 0),
) -> HelixObject: ...


def helix_obj(*args, **kwargs) -> HelixObject:
    pivot = _regularize_obj_position(kwargs.pop("pivot", (0, 0, 0)))

    error = kwargs.pop("error", None)
    if isinstance(error, ak.Array):
        error = error.to_numpy()

    # given helix parameters as positional arguments
    if len(args) > 0:
        _check_kwargs_used_up(kwargs)
        return HelixObject(*args, error=error, pivot=pivot)

    # given helix parameters as keyword arguments
    if "dr" in kwargs:
        dr = kwargs.pop("dr")
        phi0 = kwargs.pop("phi0")
        kappa = kwargs.pop("kappa")
        dz = kwargs.pop("dz")
        tanl = kwargs.pop("tanl")

        _check_kwargs_used_up(kwargs)
        return HelixObject(
            dr=dr, phi0=phi0, kappa=kappa, dz=dz, tanl=tanl, pivot=pivot, error=error
        )

    # given helix parameters as a tuple
    if "params" in kwargs:
        params = kwargs.pop("params")
        if len(params) != 5:
            raise ValueError("params must be a tuple of 5 elements")

        _check_kwargs_used_up(kwargs)
        return HelixObject(*params, pivot=pivot, error=error)

    # given momentum, position and charge
    charge: Literal[-1, 1] = int(kwargs.pop("charge"))
    assert charge in (-1, 1), "Charge must be either -1 or 1"

    momentum = _regularize_obj_momentum(kwargs.pop("momentum"))
    position = _regularize_obj_position(kwargs.pop("position"))

    # compute helix parameters
    kappa = charge / momentum.pt
    phi0 = (momentum.phi - np.pi / 2) % (2 * np.pi)

    dist = (position - pivot).to_2D()
    dr = dist.rho
    if not np.isclose(dist.phi % (2 * np.pi), phi0):
        dr *= -1

    dz = position.z - pivot.z
    tanl = momentum.pz / momentum.pt

    _check_kwargs_used_up(kwargs)
    return HelixObject(
        dr=dr,
        phi0=phi0,
        kappa=kappa,
        dz=dz,
        tanl=tanl,
        pivot=pivot,
        error=error,
    )


###############################################################################################


@nb.vectorize(cache=True)
def dr_phi0_to_x(dr: FloatLike, phi0: FloatLike) -> FloatLike:
    """
    Convert helix parameters to x location.

    Parameters:
        dr (float): helix[0] parameter, dr.
        phi0 (float): helix[1] parameter, phi0.

    Returns:
        x location of the helix.
    """
    return dr * np.cos(phi0)


@nb.vectorize(cache=True)
def dr_phi0_to_y(dr: FloatLike, phi0: FloatLike) -> FloatLike:
    """
    Convert helix parameters to y location.

    Parameters:
        dr: helix[0] parameter, dr.
        phi0: helix[1] parameter, phi0.

    Returns:
        y location of the helix.
    """
    return dr * np.sin(phi0)


@nb.vectorize(cache=True)
def phi0_to_phi(phi0: FloatLike) -> FloatLike:
    """
    Convert helix parameter phi0 to momentum phi.

    Parameters:
        phi0: helix[1] parameter, phi0.

    Returns:
        phi of the momentum vector.
    """
    return (phi0 + np.pi / 2) % (2 * np.pi)


@nb.vectorize(cache=True)
def kappa_to_pt(kappa: FloatLike) -> FloatLike:
    """
    Convert helix parameter to pt.

    Parameters:
        kappa: helix[2] parameter, kappa.

    Returns:
        pt of the helix.
    """
    return 1 / np.abs(kappa)


@nb.vectorize(cache=True)
def kappa_to_charge(kappa: IntLike) -> FloatLike:
    """
    Convert helix parameter to charge.

    Parameters:
        kappa: helix[2] parameter, kappa.

    Returns:
        charge of the helix.
    """
    return np.int8(1) if kappa > 1e-10 else np.int8(-1) if kappa < -1e-10 else np.int8(0)


@nb.vectorize(cache=True)
def kappa_to_radius(kappa: FloatLike) -> FloatLike:
    """
    Convert helix parameter kappa to circular radius.

    Parameters:
        kappa: helix[2] parameter, kappa.

    Returns:
        circular radius of the helix in cm.
    """
    return 1000 / 2.99792458 / np.abs(kappa)


###############################################################################################


def _compute_momentum(kappa, tanl, phi0):
    pt = kappa_to_pt(kappa)
    pz = pt * tanl
    phi = phi0_to_phi(phi0)
    return pt, phi, pz


def _compute_position(dr, phi0, dz):
    x = dr_phi0_to_x(dr, phi0)
    y = dr_phi0_to_y(dr, phi0)
    z = dz

    return x, y, z


###############################################################################################


class HelixAwkwardRecord(ak.Record):
    @property
    def momentum(self) -> vector.MomentumObject3D:
        """
        Returns the momentum of the helix as a 3D vector.

        Returns:
            vector.MomentumObject3D: The momentum vector of the helix.
        """
        pt, phi, pz = _compute_momentum(self.kappa, self.tanl, self.phi0)
        return ak.zip({"pt": pt, "phi": phi, "pz": pz}, with_name="Momentum3D")

    @property
    def position(self) -> vector.VectorObject3D:
        """
        Returns the position of the helix at a given azimuthal angle.

        Returns:
            vector.VectorObject3D: The position vector of the helix.
        """
        x, y, z = _compute_position(self.dr, self.phi0, self.dz)
        return ak.zip({"x": x, "y": y, "z": z}, with_name="Vector3D")

    @property
    def charge(self) -> int:
        """
        Returns the charge of the helix.

        Returns:
            int: The charge of the helix.
        """
        return kappa_to_charge(self.kappa)

    @property
    def radius(self) -> float:
        """
        Returns the radius of the helix.

        Returns:
            float: The radius of the helix in mm.
        """
        return kappa_to_radius(self.kappa)

    def change_pivot(self, *args):
        multi_trk = isinstance(self.pivot.x, ak.Array)

        # regularize pivot
        if multi_trk:
            if len(args) == 3:
                x, y, z = args
                new_pivot = None
            elif len(args) == 1:
                new_pivot = args[0]
                if isinstance(new_pivot, (vector.VectorObject3D, ak.Array, ak.Record)):
                    x = new_pivot.x
                    y = new_pivot.y
                    z = new_pivot.z
                    new_pivot = None
                else:
                    x = new_pivot[0]
                    y = new_pivot[1]
                    z = new_pivot[2]
                    new_pivot = None
            else:
                raise ValueError(
                    "change_pivot requires either 3 arguments (x, y, z) or 1 argument (pivot)."
                )

            if new_pivot is None:
                x = (
                    ak.ones_like(self.dr) * x
                    if not isinstance(x, (ak.Array, ak.Record))
                    else x
                )
                y = (
                    ak.ones_like(self.dr) * y
                    if not isinstance(y, (ak.Array, ak.Record))
                    else y
                )
                z = (
                    ak.ones_like(self.dr) * z
                    if not isinstance(z, (ak.Array, ak.Record))
                    else z
                )
                new_pivot = ak.Array({"x": x, "y": y, "z": z}, with_name="Vector3D")

            old_pivot_x = _flat_to_numpy(self.pivot.x)
            old_pivot_y = _flat_to_numpy(self.pivot.y)
            old_pivot_z = _flat_to_numpy(self.pivot.z)
            old_pivot = vector.arr({"x": old_pivot_x, "y": old_pivot_y, "z": old_pivot_z})

            new_pivot_x = _flat_to_numpy(new_pivot.x)
            new_pivot_y = _flat_to_numpy(new_pivot.y)
            new_pivot_z = _flat_to_numpy(new_pivot.z)
            new_pivot = vector.arr({"x": new_pivot_x, "y": new_pivot_y, "z": new_pivot_z})

        else:
            new_pivot = _regularize_obj_position(args)
            old_pivot = vector.obj(
                x=self.pivot.x,
                y=self.pivot.y,
                z=self.pivot.z,
            )

        # do transformation
        r = _flat_to_numpy(self.radius)
        old_dr = _flat_to_numpy(self.dr)
        old_phi0 = _flat_to_numpy(self.phi0)
        old_dz = _flat_to_numpy(self.dz)
        tanl = _flat_to_numpy(self.tanl)
        kappa = _flat_to_numpy(self.kappa)

        if "error" in self.fields:
            old_error = (
                _flat_to_numpy(self.error).reshape(-1, 5, 5)
                if multi_trk
                else _flat_to_numpy(self.error).reshape(5, 5)
            )
        else:
            old_error = None

        new_dr, new_phi0, new_dz, new_error = _change_pivot(
            r=r,
            old_dr=old_dr,
            old_phi0=old_phi0,
            old_dz=old_dz,
            kappa=kappa,
            tanl=tanl,
            old_error=old_error,
            old_pivot=old_pivot,
            new_pivot=new_pivot,
        )

        res_dict = {
            "dr": new_dr,
            "phi0": new_phi0,
            "kappa": kappa,
            "dz": new_dz,
            "tanl": tanl,
            "pivot": ak.Record(
                {"x": new_pivot.x, "y": new_pivot.y, "z": new_pivot.z},
                with_name="Vector3D",
            ),
        }

        if new_error is not None:
            res_dict["error"] = new_error

        res = ak.Record(res_dict, with_name="Bes3Helix")

        if multi_trk:
            raw_shape = _extract_index(self.dr.layout)
            for count in raw_shape:
                res = ak.unflatten(res, count)

        return res

    def isclose(
        self,
        value: "HelixAwkwardRecord",
        *,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        equal_nan: bool = False,
    ) -> bool:
        """
        Check if two helix records are close to each other.

        Args:
            value (HelixAwkwardRecord): The helix record to compare with.

        Returns:
            bool: True if the records are close, False otherwise.
        """
        if xor(
            ("error" in self.fields),
            ("error" in value.fields),
        ):
            warnings.warn(
                "One of the helix records has an error matrix, but the other does not. "
                "Ignoring error matrix for isclose check.",
                UserWarning,
            )

        multi_trk = isinstance(self.pivot.x, ak.Array)
        if multi_trk:
            return _arr_isclose(self, value, rtol=rtol, atol=atol, equal_nan=equal_nan)
        else:
            return _obj_isclose(self, value, rtol=rtol, atol=atol, equal_nan=equal_nan)


ak.behavior["Bes3Helix"] = HelixAwkwardRecord


###############################################################################################


class HelixAwkwardArray(ak.Array):
    @property
    def momentum(self) -> vec_ak.MomentumAwkward3D:
        """
        Returns the momentum of the helix as an awkward array of 3D vectors.

        Returns:
            vector.MomentumNumpy3D: The momentum vectors of the helix.
        """
        pt, phi, pz = _compute_momentum(self.kappa, self.tanl, self.phi0)
        return ak.zip({"pt": pt, "phi": phi, "pz": pz}, with_name="Momentum3D")

    @property
    def position(self) -> vec_ak.VectorAwkward3D:
        """
        Returns the position of the helix at a given azimuthal angle as an awkward array of 3D vectors.

        Returns:
            vector.VectorNumpy3D: The position vectors of the helix.
        """
        x, y, z = _compute_position(self.dr, self.phi0, self.dz)
        return ak.zip({"x": x, "y": y, "z": z}, with_name="Vector3D")

    @property
    def charge(self) -> ak.Array:
        """
        Returns the charge of the helix as an awkward array.

        Returns:
            ak.Array: The charge of the helix.
        """
        return kappa_to_charge(self.kappa)

    @property
    def radius(self) -> ak.Array:
        """
        Returns the radius of the helix as an awkward array.

        Returns:
            ak.Array: The radius of the helix in mm.
        """
        return kappa_to_radius(self.kappa)

    def change_pivot(self, *args) -> "HelixAwkwardArray":
        """
        Changes the pivot point of the helix.
        """
        if len(args) == 3:
            x, y, z = args
            pivot = None
        elif len(args) == 1:
            pivot = args[0]
            if isinstance(pivot, vector.VectorObject3D):
                x = pivot.x
                y = pivot.y
                z = pivot.z
                pivot = None
            elif not isinstance(pivot, ak.Array):
                x = pivot[0]
                y = pivot[1]
                z = pivot[2]
                pivot = None
        else:
            raise ValueError(
                "change_pivot requires either 3 arguments (x, y, z) or 1 argument (pivot)."
            )

        if pivot is None:
            x = ak.ones_like(self.dr) * x if not isinstance(x, ak.Array) else x
            y = ak.ones_like(self.dr) * y if not isinstance(y, ak.Array) else y
            z = ak.ones_like(self.dr) * z if not isinstance(z, ak.Array) else z
            pivot = ak.Array({"x": x, "y": y, "z": z}, with_name="Vector3D")

        raw_shape = _extract_index(self.dr.layout)

        r = _flat_to_numpy(self.radius)

        old_dr = _flat_to_numpy(self.dr)
        old_phi0 = _flat_to_numpy(self.phi0)
        old_dz = _flat_to_numpy(self.dz)
        tanl = _flat_to_numpy(self.tanl)
        kappa = _flat_to_numpy(self.kappa)

        old_pivot_x = _flat_to_numpy(self.pivot.x)
        old_pivot_y = _flat_to_numpy(self.pivot.y)
        old_pivot_z = _flat_to_numpy(self.pivot.z)
        old_pivot = vector.arr({"x": old_pivot_x, "y": old_pivot_y, "z": old_pivot_z})

        new_pivot_x = _flat_to_numpy(pivot.x)
        new_pivot_y = _flat_to_numpy(pivot.y)
        new_pivot_z = _flat_to_numpy(pivot.z)
        new_pivot = vector.arr({"x": new_pivot_x, "y": new_pivot_y, "z": new_pivot_z})

        old_error = (
            _flat_to_numpy(self.error).reshape(-1, 5, 5) if "error" in self.fields else None
        )

        new_dr, new_phi0, new_dz, new_error = _change_pivot(
            r=r,
            old_dr=old_dr,
            old_phi0=old_phi0,
            old_dz=old_dz,
            kappa=kappa,
            tanl=tanl,
            old_error=old_error,
            old_pivot=old_pivot,
            new_pivot=new_pivot,
        )

        res_dict = {
            "dr": new_dr,
            "phi0": new_phi0,
            "kappa": kappa,
            "tanl": tanl,
            "dz": new_dz,
            "pivot": ak.zip(
                {"x": new_pivot.x, "y": new_pivot.y, "z": new_pivot.z},
                with_name="Vector3D",
            ),
        }

        if new_error is not None:
            res_dict["error"] = new_error

        raw_shape = _extract_index(self.dr.layout)
        res = ak.Array(res_dict, with_name="Bes3Helix")

        for count in raw_shape:
            res = ak.unflatten(res, count)

        return res

    def isclose(
        self,
        other: "HelixAwkwardRecord",
        *,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        equal_nan: bool = False,
    ) -> ak.Array:
        """
        Check if two helix records are close to each other.

        Args:
            other (HelixAwkwardRecord): The helix record to compare with.

        Returns:
            ak.Array: A boolean array indicating if the records are close.
        """
        if xor(
            ("error" in self.fields),
            ("error" in other.fields),
        ):
            warnings.warn(
                "One of the helix records has an error matrix, but the other does not. "
                "Ignoring error matrix for isclose check.",
                UserWarning,
            )

        return _arr_isclose(self, other, rtol=rtol, atol=atol, equal_nan=equal_nan)


ak.behavior["*", "Bes3Helix"] = HelixAwkwardArray

###############################################################################################


# case 1
@overload
def helix_awk(
    helix: ak.Array,
    error: Optional[ak.Array] = None,
    pivot: TypeAwkPosition = (0, 0, 0),
) -> HelixAwkwardArray: ...


# case 2
@overload
def helix_awk(
    *,
    dr: ak.Array,
    phi0: ak.Array,
    kappa: ak.Array,
    dz: ak.Array,
    tanl: ak.Array,
    error: Optional[ak.Array] = None,
    pivot: TypeAwkPosition = (0, 0, 0),
) -> HelixAwkwardArray: ...


# case 3
@overload
def helix_awk(
    *,
    momentum: ak.Array,
    position: ak.Array,
    charge: Union[Literal[-1, 1], ak.Array],
    error: Optional[ak.Array] = None,
    pivot: TypeAwkPosition = (0, 0, 0),
) -> HelixAwkwardArray: ...


@nb.vectorize(cache=True)
def _fix_dr_sign(dr: FloatLike, phi0: FloatLike, dist_phi: FloatLike) -> FloatLike:
    """
    Fix the sign of dr based on the azimuthal angle.

    Parameters:
        dr (float): The radial distance.
        phi0 (float): The azimuthal angle.
        dist_phi (float): The difference between the azimuthal angle and phi0.

    Returns:
        float: The corrected radial distance.
    """
    if not np.isclose(dist_phi % (2 * np.pi), phi0):
        return -dr
    return dr


def helix_awk(*args, **kwargs) -> HelixAwkwardArray:
    error = None
    pivot = (0, 0, 0)

    if len(args) > 0:
        helix = args[0]

        if len(args) > 1:
            error = args[1]
            if "error" in kwargs:
                raise ValueError("Cannot pass both helix and error as positional arguments.")

        if len(args) > 2:
            pivot = args[2]
            if "pivot" in kwargs:
                raise ValueError("Cannot pass both helix and pivot as positional arguments.")

        dr = helix[..., 0]
        phi0 = helix[..., 1]
        kappa = helix[..., 2]
        dz = helix[..., 3]
        tanl = helix[..., 4]

    elif "helix" in kwargs:
        helix = kwargs.pop("helix")

        dr = helix[..., 0]
        phi0 = helix[..., 1]
        kappa = helix[..., 2]
        dz = helix[..., 3]
        tanl = helix[..., 4]

    elif "dr" in kwargs:
        dr = kwargs.pop("dr")
        phi0 = kwargs.pop("phi0")
        kappa = kwargs.pop("kappa")
        dz = kwargs.pop("dz")
        tanl = kwargs.pop("tanl")

    else:
        momentum = kwargs.pop("momentum")
        position = kwargs.pop("position")
        charge = kwargs.pop("charge")

        pivot = kwargs.pop("pivot", (0, 0, 0))
        if not isinstance(pivot, ak.Array):
            pivot = _regularize_obj_position(pivot)

        # compute helix parameters
        kappa = charge / momentum.pt
        phi0 = (momentum.phi - np.pi / 2) % (2 * np.pi)

        dist = (position - pivot).to_2D()
        dr = _fix_dr_sign(dist.rho, phi0, dist.phi)

        dz = position.z - pivot.z
        tanl = momentum.pz / momentum.pt

    error = kwargs.pop("error", error)
    pivot = kwargs.pop("pivot", pivot)

    if not isinstance(pivot, ak.Array):
        pivot = _regularize_obj_position(pivot)

        x0 = ak.ones_like(dr) * pivot.x
        y0 = ak.ones_like(dr) * pivot.y
        z0 = ak.ones_like(dr) * pivot.z
        pivot = ak.zip({"x": x0, "y": y0, "z": z0}, with_name="Vector3D")

    res_dict = {
        "dr": dr,
        "phi0": phi0,
        "kappa": kappa,
        "dz": dz,
        "tanl": tanl,
        "pivot": pivot,
    }

    if error is not None:
        res_dict["error"] = error

    _check_kwargs_used_up(kwargs)

    raw_shape = _extract_index(dr.layout)
    return ak.zip(res_dict, depth_limit=len(raw_shape) + 1, with_name="Bes3Helix")



import warnings
from typing import Literal, Union

import awkward as ak
import numba as nb
import numpy as np

from ..typing import FloatLike, IntLike


@nb.vectorize(cache=True)
def _helix01_to_x(helix0: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert helix parameters to x location.

    Parameters:
        helix0: helix[0] parameter, dr.
        helix1: helix[1] parameter, phi0.

    Returns:
        x location of the helix.
    """
    return helix0 * np.cos(helix1)


@nb.vectorize(cache=True)
def _helix01_to_y(helix0: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert helix parameters to y location.

    Parameters:
        helix0: helix[0] parameter, dr.
        helix1: helix[1] parameter, phi0.

    Returns:
        y location of the helix.
    """
    return helix0 * np.sin(helix1)


@nb.vectorize(cache=True)
def _helix2_to_pt(
    helix2: FloatLike,
) -> FloatLike:
    """
    Convert helix parameter to pt.

    Parameters:
        helix2: helix[2] parameter, kappa.

    Returns:
        pt of the helix.
    """
    return 1 / np.abs(helix2)


@nb.vectorize(cache=True)
def _helix2_to_charge(
    helix2: FloatLike,
) -> FloatLike:
    """
    Convert helix parameter to charge.

    Parameters:
        helix2: helix[2] parameter, kappa.

    Returns:
        charge of the helix.
    """
    if -1e-10 < helix2 < 1e-10:
        return np.int8(0)
    return np.int8(1) if helix2 > 0 else np.int8(-1)


@nb.vectorize(cache=True)
def _pt_helix1_to_px(pt: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert pt and helix1 to px.

    Parameters:
        pt: pt of the helix.
        helix1: helix[1] parameter, phi0.

    Returns:
        px of the helix.
    """
    return -pt * np.sin(helix1)


@nb.vectorize(cache=True)
def _pt_helix1_to_py(pt: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert pt and helix1 to py.

    Parameters:
        pt: pt of the helix.
        helix1: helix[1] parameter, phi0.

    Returns:
        py of the helix.
    """
    return pt * np.cos(helix1)


@nb.vectorize(cache=True)
def _pt_helix4_to_p(pt: FloatLike, helix4: FloatLike) -> FloatLike:
    """
    Convert pt and helix4 to p.

    Parameters:
        pt: pt of the helix.
        helix4: helix[4] parameter, tanl.

    Returns:
        p of the helix.
    """
    return pt * np.sqrt(1 + helix4**2)


@nb.vectorize(cache=True)
def _pz_p_to_theta(pz: FloatLike, p: FloatLike) -> FloatLike:
    """
    Convert pz and p to theta.

    Parameters:
        pz: pz of the helix.
        p: p of the helix.

    Returns:
        theta of the helix.
    """
    return np.arccos(pz / p)


def parse_helix(
    helix: Union[ak.Array, np.ndarray],
    library: Literal["ak", "auto"] = "auto",
) -> Union[ak.Array, dict[str, np.ndarray]]:
    """
    Parse helix parameters to physical parameters.

    Parameters:
        helix: helix parameters, the last dimension should be 5.
        library: the library to use, if "auto", return a dict when input is np.ndarray, \
            return an ak.Array when input is ak.Array. If "ak", return an ak.Array.

    Returns:
        parsed physical parameters. "x", "y", "z", "r" for position, \
            "pt", "px", "py", "pz", "p", "theta", "phi" for momentum, \
            "charge" for charge, "r_trk" for track radius.
    """

    warnings.warn(
        "'parse_helix' is deprecated and will be removed in the future, "
        "use 'helix_obj' and 'helix_awk' instead.",
        DeprecationWarning,
    )

    helix0 = helix[..., 0]
    helix1 = helix[..., 1]
    helix2 = helix[..., 2]
    helix3 = helix[..., 3]
    helix4 = helix[..., 4]

    x = _helix01_to_x(helix0, helix1)
    y = _helix01_to_y(helix0, helix1)
    z = helix3
    r = np.abs(helix0)

    pt = _helix2_to_pt(helix2)
    px = _pt_helix1_to_px(pt, helix1)
    py = _pt_helix1_to_py(pt, helix1)
    pz = pt * helix4
    p = _pt_helix4_to_p(pt, helix4)
    theta = _pz_p_to_theta(pz, p)
    phi = np.arctan2(py, px)

    charge = _helix2_to_charge(helix2)

    r_trk = pt * (10 / 2.99792458)  # |pt| * [GeV/c] / 1[e] / 1[T] = |pt| * 10/3 [m]
    r_trk = r_trk * 100  # to [cm]

    res_dict = {
        "x": x,
        "y": y,
        "z": z,
        "r": r,
        "px": px,
        "py": py,
        "pz": pz,
        "pt": pt,
        "p": p,
        "theta": theta,
        "phi": phi,
        "charge": charge,
        "r_trk": r_trk,
    }

    if isinstance(helix, ak.Array) or library == "ak":
        return ak.zip(res_dict)
    else:
        return res_dict


@nb.vectorize(cache=True)
def _xy_to_helix0(x: FloatLike, y: FloatLike) -> FloatLike:
    """
    Convert x and y to helix[0] parameter, dr.

    Parameters:
        x: x location of the helix.
        y: y location of the helix.

    Returns:
        helix[0] parameter, dr.
    """
    return np.sqrt(x**2 + y**2)


@nb.vectorize(cache=True)
def _xy_to_helix1(x: FloatLike, y: FloatLike) -> FloatLike:
    """
    Convert x and y to helix[1] parameter, phi0.

    Parameters:
        x: x location of the helix.
        y: y location of the helix.

    Returns:
        helix[1] parameter, phi0.
    """
    return (np.arctan2(y, x) + 2 * np.pi) % (2 * np.pi)  # make sure that phi0 is in [0, 2pi)


@nb.vectorize(cache=True)
def _ptcharge_to_helix2(pt: FloatLike, charge: IntLike) -> FloatLike:
    """
    Convert pt, charge to helix[2] parameter, kappa.

    Parameters:
        pt: pt of the helix.
        charge: charge of the helix.

    Returns:
        helix[2] parameter, kappa.
    """
    return np.sign(charge) / pt if pt != 0 else 0


@nb.vectorize(cache=True)
def _pxpy_to_pt(px: FloatLike, py: FloatLike) -> FloatLike:
    """
    Convert px and py to pt.

    Parameters:
        px: px of the helix.
        py: py of the helix.

    Returns:
        pt of the helix.
    """
    return np.sqrt(px**2 + py**2)


def compute_helix(
    data: Union[ak.Array, dict[str, np.ndarray]],
    library: Literal["ak", "auto"] = "auto",
) -> Union[ak.Array, np.ndarray]:
    """
    Compute helix parameters from charge, position and momentum.

    Parameters:
        data: data to compute helix parameters from, should contain "x", "y", "z", "px", \
            "py", "pz" and "charge" keys.
        library: the library to use, if "auto", return an np.ndarray when input is a dictionary \
            of np.ndarray, return an ak.Array when input is ak.Array. If "ak", return an ak.Array.

    Returns:
        helix parameters, the shape is (..., 5), where the last dimension is (dr, phi0, kappa, z, tanl).
    """
    warnings.warn(
        "'compute_helix' is deprecated and will be removed in the future, "
        "use 'helix_obj' and 'helix_awk' instead.",
        DeprecationWarning,
    )

    x = data["x"]
    y = data["y"]
    z = data["z"]
    px = data["px"]
    py = data["py"]
    pz = data["pz"]
    charge = data["charge"]

    pt = _pxpy_to_pt(px, py)

    helix0 = _xy_to_helix0(x, y)
    helix1 = _xy_to_helix1(x, y)
    helix2 = _ptcharge_to_helix2(pt, charge)
    helix3 = z
    helix4 = pz / pt

    if isinstance(data, ak.Array) or library == "ak":
        return ak.concatenate(
            [
                ak.unflatten(helix0, 1, axis=-1),
                ak.unflatten(helix1, 1, axis=-1),
                ak.unflatten(helix2, 1, axis=-1),
                ak.unflatten(helix3, 1, axis=-1),
                ak.unflatten(helix4, 1, axis=-1),
            ],
            axis=-1,
        )  # make the shape be (..., 5)
    else:
        return np.vstack([helix0, helix1, helix2, helix3, helix4]).T


@nb.vectorize(cache=True)
def _regularize_phi0(dr: FloatLike, phi0: FloatLike) -> FloatLike:
    """
    Regularize phi0 to make sure that dr is positive.

    Parameters:
        dr: helix[0] parameter, dr.
        phi0: helix[1] parameter, phi0.

    Returns:
        phi0: helix[1] parameter, phi0. Always in [0, 2pi).
    """
    if dr < 0:
        phi0 += 3 * np.pi
    return phi0 % (2 * np.pi)  # make sure that phi0 is in [0, 2pi)


def regularize_helix(helix: Union[ak.Array, np.ndarray]) -> Union[ak.Array, np.ndarray]:
    """
    Sometimes, the `dr` may be negative, which is not allowed. This function will make sure \
    that the `dr` is positive by rotating `phi0`.

    Parameters:
        helix: helix parameters, the last dimension should be 5.

    Returns:
        helix: helix parameters, the last dimension should be 5, and the `dr` is positive.
        The `phi0` is rotated to make sure that the `dr` is positive.
    """

    warnings.warn(
        "'regularize_helix' is deprecated and will be removed in the future, "
        "use 'helix_obj' and 'helix_awk' instead.",
        DeprecationWarning,
    )

    helix0 = helix[..., 0]
    helix1 = helix[..., 1]
    helix2 = helix[..., 2]
    helix3 = helix[..., 3]
    helix4 = helix[..., 4]

    helix1 = _regularize_phi0(helix0, helix1)
    helix0 = np.abs(helix0)

    if isinstance(helix, ak.Array):
        return ak.concatenate(
            [
                ak.unflatten(helix0, 1, axis=-1),
                ak.unflatten(helix1, 1, axis=-1),
                ak.unflatten(helix2, 1, axis=-1),
                ak.unflatten(helix3, 1, axis=-1),
                ak.unflatten(helix4, 1, axis=-1),
            ],
            axis=-1,
        )
    else:
        return np.vstack([helix0, helix1, helix2, helix3, helix4]).T



import importlib
import os
from pathlib import Path

import pybes3 as p3


def test_detector_geometry_mdc(capsys, monkeypatch):
    monkeypatch.setenv("PYBES3_NUMBA_CACHE_SILENT", "1")

    geom_dir = Path(p3.detectors.geometry.__file__).parent

    # Change mdc_geom.npz mtime
    mdc_geom_path = geom_dir / "mdc_geom.npz"
    os.utime(mdc_geom_path)

    # Reload pybes3
    importlib.reload(p3)

    # Check output
    captured = capsys.readouterr()
    assert captured.out.startswith("Removed cache files:")


def test_detector_geometry_emc(capsys, monkeypatch):
    monkeypatch.setenv("PYBES3_NUMBA_CACHE_SILENT", "1")

    geom_dir = Path(p3.detectors.geometry.__file__).parent

    # Change emc_geom.npz mtime
    emc_geom_path = geom_dir / "emc_geom.npz"
    os.utime(emc_geom_path)

    # Reload pybes3
    importlib.reload(p3)

    # Check output
    captured = capsys.readouterr()
    assert captured.out.startswith("Removed cache files:")


if __name__ == "__main__":
    import pytest

    pytest.main([__file__, "-v", "-s"])



import re
from enum import Enum
from typing import Literal, Union

import awkward as ak
import awkward.contents
import awkward.index
import numpy as np
import uproot
import uproot.behaviors.TBranch
import uproot.extras
import uproot.interpretation

from .._utils import _extract_index, _flat_to_numpy, _recover_shape
from . import besio_cpp as bcpp

type_np2array = {
    "u1": "B",
    "u2": "H",
    "u4": "I",
    "u8": "Q",
    "i1": "b",
    "i2": "h",
    "i4": "i",
    "i8": "q",
    "f": "f",
    "d": "d",
}

num_typenames = {
    "bool": "i1",
    "char": "i1",
    "short": "i2",
    "int": "i4",
    "long": "i8",
    "unsigned char": "u1",
    "unsigned short": "u2",
    "unsigned int": "u4",
    "unsigned long": "u8",
    "float": "f",
    "double": "d",
    # cstdint
    "int8_t": "i1",
    "int16_t": "i2",
    "int32_t": "i4",
    "int64_t": "i8",
    "uint8_t": "u1",
    "uint16_t": "u2",
    "uint32_t": "u4",
    "uint64_t": "u8",
    # ROOT types
    "Bool_t": "i1",
    "Char_t": "i1",
    "Short_t": "i2",
    "Int_t": "i4",
    "Long_t": "i8",
    "UChar_t": "u1",
    "UShort_t": "u2",
    "UInt_t": "u4",
    "ULong_t": "u8",
    "Float_t": "f",
    "Double_t": "d",
}

stl_typenames = {
    "vector",
    "array",
    "map",
    "unordered_map",
    "string",
    "multimap",
}


tarray_typenames = {
    "TArrayC": "i1",
    "TArrayS": "i2",
    "TArrayI": "i4",
    "TArrayL": "i8",
    "TArrayF": "f",
    "TArrayD": "d",
}

readers: set["BaseReader"] = set()


ctype_hints = Literal["bool", "i1", "i2", "i4", "i8", "u1", "u2", "u4", "u8", "f", "d"]


class ReaderType(Enum):
    CType = "CType"
    STLSequence = "STLSequence"
    STLMap = "STLMap"
    STLString = "STLString"
    TArray = "TArray"
    TString = "TString"
    TObject = "TObject"
    CArray = "CArray"
    BaseObject = "BaseObject"
    ObjectHeader = "ObjectHeader"
    Empty = "Empty"


def get_top_type_name(type_name: str) -> str:
    if type_name.endswith("*"):
        type_name = type_name[:-1].strip()
    type_name = type_name.replace("std::", "").strip()
    return type_name.split("<")[0]


def gen_tree_config(
    cls_streamer_info: dict,
    all_streamer_info: dict,
    item_path: str = "",
) -> dict:
    """
    Generate reader configuration for a class streamer information.

    The content it returns should be:

    ```python
    {
        "reader": ReaderType,
        "name": str,
        "ctype": str, # for CTypeReader, TArrayReader
        "element_reader": dict, # reader config of the element, for STLVectorReader, SimpleCArrayReader, TObjectCArrayReader
        "flat_size": int, # for SimpleCArrayReader, TObjectCArrayReader
        "fMaxIndex": list[int], # for SimpleCArrayReader, TObjectCArrayReader
        "fArrayDim": int, # for SimpleCArrayReader, TObjectCArrayReader
        "key_reader": dict, # reader config of the key, for STLMapReader
        "val_reader": dict, # reader config of the value, for STLMapReader
        "sub_readers": list[dict], # for BaseObjectReader, ObjectHeaderReader
        "is_top_level": bool, # for STLVectorReader, STLMapReader, STLStringReader
    }
    ```

    Args:
        cls_streamer_info (dict): Class streamer information.
        all_streamer_info (dict): All streamer information.
        item_path (str): Path to the item.

    Returns:
        dict: Reader configuration.
    """
    fName = cls_streamer_info["fName"]
    item_path = fName if item_path == "" else f"{item_path}.{fName}"

    for reader in sorted(readers, key=lambda x: x.priority(), reverse=True):
        top_type_name = get_top_type_name(cls_streamer_info["fTypeName"])
        tree_config = reader.gen_tree_config(
            top_type_name,
            cls_streamer_info,
            all_streamer_info,
            item_path,
        )
        if tree_config is not None:
            return tree_config

    raise ValueError(f"Unknown type: {cls_streamer_info['fTypeName']} for {item_path}")


def get_reader_instance(tree_config: dict):
    for cls_reader in sorted(readers, key=lambda x: x.priority(), reverse=True):
        reader = cls_reader.get_reader_instance(tree_config)
        if reader is not None:
            return reader

    raise ValueError(f"Unknown reader type: {tree_config['reader']} for {tree_config['name']}")


def reconstruct_array(
    raw_data: Union[np.ndarray, tuple, list, None],
    tree_config: dict,
) -> Union[ak.Array, None]:
    for reader in sorted(readers, key=lambda x: x.priority(), reverse=True):
        data = reader.reconstruct_array(raw_data, tree_config)
        if data is not None:
            return data

    raise ValueError(f"Unknown reader type: {tree_config['reader']} for {tree_config['name']}")


def gen_tree_config_from_type_name(
    type_name: str,
    all_streamer_info: dict,
    item_path: str = "",
):
    return gen_tree_config(
        {
            "fName": type_name,
            "fTypeName": type_name,
        },
        all_streamer_info,
        item_path,
    )


def regularize_object_path(object_path: str) -> str:
    return re.sub(r";[0-9]+", r"", object_path)


class BaseReader:
    """
    Base class for all readers.
    """

    @classmethod
    def priority(cls) -> int:
        """
        The priority of the reader. Higher priority means the reader will be
        used first.
        """
        return 20

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name: str,
        cls_streamer_info: dict,
        all_streamer_info: dict,
        item_path: str = "",
    ) -> dict:
        raise NotImplementedError("This method should be implemented in subclasses.")

    @classmethod
    def get_reader_instance(cls, tree_config: dict) -> bcpp.BaseReader:
        """
        Args:
            tree_config (dict): The configuration dictionary for the reader.

        Returns:
            An instance of the appropriate reader class.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")

    @classmethod
    def reconstruct_array(
        cls,
        raw_data: Union[np.ndarray, tuple, list, None],
        tree_config: dict,
    ) -> Union[ak.Array, None]:
        """
        Args:
            raw_data (Union[np.ndarray, tuple, list, None]): The raw data to be
                recovered.
            tree_config (dict): The configuration dictionary for the reader.

        Returns:
            awkward.Array: The recovered data as an awkward array.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")


class CTypeReader(BaseReader):
    """
    This class reads C++ primitive types from a binary parser.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name in num_typenames:
            ctype = num_typenames[top_type_name]
            return {
                "reader": ReaderType.CType,
                "name": cls_streamer_info["fName"],
                "ctype": ctype,
            }
        else:
            return None

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.CType:
            return None

        ctype = tree_config["ctype"]
        return {
            "i1": bcpp.Int8Reader,
            "i2": bcpp.Int16Reader,
            "i4": bcpp.Int32Reader,
            "i8": bcpp.Int64Reader,
            "u1": bcpp.UInt8Reader,
            "u2": bcpp.UInt16Reader,
            "u4": bcpp.UInt32Reader,
            "u8": bcpp.UInt64Reader,
            "f": bcpp.FloatReader,
            "d": bcpp.DoubleReader,
        }[ctype](tree_config["name"])

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.CType:
            return None

        return ak.Array(raw_data)


class STLSequenceReader(BaseReader):
    """
    This class reads STL sequence (vector, array) from a binary parser.
    """

    @staticmethod
    def get_sequence_element_typename(type_name: str) -> str:
        """
        Get the element type name of a vector type.

        e.g. vector<vector<int>> -> vector<int>
        """
        type_name = (
            type_name.replace("std::", "").replace("< ", "<").replace(" >", ">").strip()
        )
        return re.match(r"^(vector|array)<(.*)>$", type_name).group(2)

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name not in ["vector", "array"]:
            return None

        fName = cls_streamer_info["fName"]
        fTypeName = cls_streamer_info["fTypeName"]
        element_type = cls.get_sequence_element_typename(fTypeName)
        element_info = {
            "fName": fName,
            "fTypeName": element_type,
        }

        element_tree_config = gen_tree_config(
            element_info,
            all_streamer_info,
            item_path,
        )

        top_element_type = get_top_type_name(element_type)
        if top_element_type in stl_typenames:
            element_tree_config["is_top"] = False

        return {
            "reader": ReaderType.STLSequence,
            "name": fName,
            "element_reader": element_tree_config,
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.STLSequence:
            return None

        element_reader = get_reader_instance(tree_config["element_reader"])
        is_top = tree_config.get("is_top", True)
        return bcpp.STLSeqReader(tree_config["name"], is_top, element_reader)

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.STLSequence:
            return None

        counts, element_raw_data = raw_data
        element_data = reconstruct_array(
            element_raw_data,
            tree_config["element_reader"],
        )
        return ak.unflatten(element_data, counts)


class STLMapReader(BaseReader):
    """
    This class reads std::map from a binary parser.
    """

    @staticmethod
    def get_map_key_val_typenames(type_name: str) -> tuple[str, str]:
        """
        Get the key and value type names of a map type.

        e.g. map<int, vector<int>> -> (int, vector<int>)
        """
        type_name = (
            type_name.replace("std::", "").replace("< ", "<").replace(" >", ">").strip()
        )
        return re.match(r"^(map|unordered_map|multimap)<(.*),(.*)>$", type_name).groups()[1:3]

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name not in ["map", "unordered_map", "multimap"]:
            return None

        fTypeName = cls_streamer_info["fTypeName"]
        key_type_name, val_type_name = cls.get_map_key_val_typenames(fTypeName)

        fName = cls_streamer_info["fName"]
        key_info = {
            "fName": "key",
            "fTypeName": key_type_name,
        }

        val_info = {
            "fName": "val",
            "fTypeName": val_type_name,
        }

        key_tree_config = gen_tree_config(key_info, all_streamer_info, item_path)
        if get_top_type_name(key_type_name) in stl_typenames:
            key_tree_config["is_top"] = False

        val_tree_config = gen_tree_config(val_info, all_streamer_info, item_path)
        if get_top_type_name(val_type_name) in stl_typenames:
            val_tree_config["is_top"] = False

        return {
            "reader": ReaderType.STLMap,
            "name": fName,
            "key_reader": key_tree_config,
            "val_reader": val_tree_config,
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.STLMap:
            return None

        key_cpp_reader = get_reader_instance(tree_config["key_reader"])
        val_cpp_reader = get_reader_instance(tree_config["val_reader"])
        is_top = tree_config.get("is_top", True)
        return bcpp.STLMapReader(
            tree_config["name"],
            is_top,
            key_cpp_reader,
            val_cpp_reader,
        )

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.STLMap:
            return None

        key_tree_config = tree_config["key_reader"]
        val_tree_config = tree_config["val_reader"]
        counts, key_raw_data, val_raw_data = raw_data
        key_data = reconstruct_array(key_raw_data, key_tree_config)
        val_data = reconstruct_array(val_raw_data, val_tree_config)

        return ak.unflatten(
            ak.zip(
                {
                    key_tree_config["name"]: key_data,
                    val_tree_config["name"]: val_data,
                },
                with_name="pair",
            ),
            counts,
        )


class STLStringReader(BaseReader):
    """
    This class reads std::string from a binary parser.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name != "string":
            return None

        return {
            "reader": ReaderType.STLString,
            "name": cls_streamer_info["fName"],
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.STLString:
            return None

        return bcpp.STLStringReader(
            tree_config["name"],
            tree_config.get("is_top", True),
        )

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.STLString:
            return None

        counts, data = raw_data
        return ak.enforce_type(ak.unflatten(data, counts), "string")


class TArrayReader(BaseReader):
    """
    This class reads TArray from a binary paerser.

    TArray includes TArrayC, TArrayS, TArrayI, TArrayL, TArrayF, and TArrayD.
    Corresponding ctype is u1, u2, i4, i8, f, and d.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name not in tarray_typenames:
            return None

        ctype = tarray_typenames[top_type_name]
        return {
            "reader": ReaderType.TArray,
            "name": cls_streamer_info["fName"],
            "ctype": ctype,
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.TArray:
            return None

        ctype = tree_config["ctype"]

        return {
            "i1": bcpp.TArrayCReader,
            "i2": bcpp.TArraySReader,
            "i4": bcpp.TArrayIReader,
            "i8": bcpp.TArrayLReader,
            "f": bcpp.TArrayFReader,
            "d": bcpp.TArrayDReader,
        }[ctype](tree_config["name"])

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.TArray:
            return None

        counts, data = raw_data
        return ak.unflatten(data, counts)


class TStringReader(BaseReader):
    """
    This class reads TString from a binary parser.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name != "TString":
            return None

        return {
            "reader": ReaderType.TString,
            "name": cls_streamer_info["fName"],
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.TString:
            return None

        return bcpp.TStringReader(tree_config["name"])

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.TString:
            return None

        counts, data = raw_data
        offsets = np.zeros(len(counts) + 1, dtype=counts.dtype)
        np.cumsum(counts, out=offsets[1:])

        return ak.Array(
            awkward.contents.ListOffsetArray(
                awkward.index.Index(offsets),
                awkward.contents.NumpyArray(data, parameters={"__array__": "char"}),
                parameters={"__array__": "string"},
            )
        )


class TObjectReader(BaseReader):
    """
    This class reads TObject from a binary parser.

    It will not record any data.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name != "BASE":
            return None

        fType = cls_streamer_info["fType"]
        if fType != 66:
            return None

        return {
            "reader": ReaderType.TObject,
            "name": cls_streamer_info["fName"],
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.TObject:
            return None

        return bcpp.TObjectReader(tree_config["name"])

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        return None


class CArrayReader(BaseReader):
    """
    This class reads a C-array from a binary parser.
    """

    @classmethod
    def priority(cls):
        return 100  # This reader should be called first

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if cls_streamer_info.get("fArrayDim", 0) == 0:
            return None

        fName = cls_streamer_info["fName"]
        fTypeName = cls_streamer_info["fTypeName"]
        fArrayDim = cls_streamer_info["fArrayDim"]
        fMaxIndex = cls_streamer_info["fMaxIndex"]

        element_streamer_info = cls_streamer_info.copy()
        element_streamer_info["fArrayDim"] = 0

        element_tree_config = gen_tree_config(
            element_streamer_info,
            all_streamer_info,
        )

        flat_size = np.prod(fMaxIndex[:fArrayDim])
        assert flat_size > 0, f"flatten_size should be greater than 0, but got {flat_size}"

        # c-type number or TArray
        if top_type_name in num_typenames or top_type_name in tarray_typenames:
            return {
                "reader": ReaderType.CArray,
                "name": fName,
                "is_obj": False,
                "element_reader": element_tree_config,
                "flat_size": flat_size,
                "fMaxIndex": fMaxIndex,
                "fArrayDim": fArrayDim,
            }

        # TSTring
        elif top_type_name == "TString":
            return {
                "reader": ReaderType.CArray,
                "name": fName,
                "is_obj": True,
                "element_reader": element_tree_config,
                "flat_size": flat_size,
                "fMaxIndex": fMaxIndex,
                "fArrayDim": fArrayDim,
            }

        # STL
        elif top_type_name in stl_typenames:
            element_tree_config["is_top"] = False
            return {
                "reader": ReaderType.CArray,
                "name": fName,
                "is_obj": True,
                "flat_size": flat_size,
                "element_reader": element_tree_config,
                "fMaxIndex": fMaxIndex,
                "fArrayDim": fArrayDim,
            }

        else:
            raise ValueError(f"Unknown type: {top_type_name} for C-array: {fTypeName}")

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        reader_type = tree_config["reader"]
        if reader_type != ReaderType.CArray:
            return None

        element_reader = get_reader_instance(tree_config["element_reader"])

        return bcpp.CArrayReader(
            tree_config["name"],
            tree_config["is_obj"],
            tree_config["flat_size"],
            element_reader,
        )

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.CArray:
            return None

        element_tree_config = tree_config["element_reader"]
        fMaxIndex = tree_config["fMaxIndex"]
        fArrayDim = tree_config["fArrayDim"]
        shape = [fMaxIndex[i] for i in range(fArrayDim)]

        element_data = reconstruct_array(
            raw_data,
            element_tree_config,
        )

        for s in shape[::-1]:
            element_data = ak.unflatten(element_data, s)

        return element_data


class BaseObjectReader(BaseReader):
    """
    Base class is what a custom class inherits from.
    It has fNBytes(uint32), fVersion(uint16) at the beginning.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        if top_type_name != "BASE":
            return None

        fType = cls_streamer_info["fType"]
        if fType != 0:
            return None

        fName = cls_streamer_info["fName"]
        sub_streamers: list = all_streamer_info[fName]

        sub_tree_configs = [
            gen_tree_config(s, all_streamer_info, item_path) for s in sub_streamers
        ]

        return {
            "reader": ReaderType.BaseObject,
            "name": fName,
            "sub_readers": sub_tree_configs,
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.BaseObject:
            return None

        sub_readers = [get_reader_instance(s) for s in tree_config["sub_readers"]]
        return bcpp.BaseObjectReader(tree_config["name"], sub_readers)

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.BaseObject:
            return None

        sub_tree_configs = tree_config["sub_readers"]

        arr_dict = {}
        for s_cfg, s_data in zip(sub_tree_configs, raw_data):
            s_name = s_cfg["name"]
            s_reader_type = s_cfg["reader"]

            if s_reader_type == ReaderType.TObject:
                continue

            arr_dict[s_name] = reconstruct_array(s_data, s_cfg)

        return ak.Array(arr_dict)


class ObjectHeaderReader(BaseReader):
    """
    This class read an object starting with an object header.
    """

    @classmethod
    def priority(cls):
        return 0  # should be called last

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        sub_streamers: list = all_streamer_info[top_type_name]
        sub_tree_configs = [
            gen_tree_config(s, all_streamer_info, item_path) for s in sub_streamers
        ]
        return {
            "reader": ReaderType.ObjectHeader,
            "name": top_type_name,
            "sub_readers": sub_tree_configs,
        }

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.ObjectHeader:
            return None

        sub_readers = [get_reader_instance(s) for s in tree_config["sub_readers"]]
        return bcpp.ObjectHeaderReader(tree_config["name"], sub_readers)

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.ObjectHeader:
            return None

        sub_tree_configs = tree_config["sub_readers"]

        arr_dict = {}
        for s_cfg, s_data in zip(sub_tree_configs, raw_data):
            s_name = s_cfg["name"]
            s_reader_type = s_cfg["reader"]

            if s_reader_type == ReaderType.TObject:
                continue

            arr_dict[s_name] = reconstruct_array(s_data, s_cfg)

        return ak.Array(arr_dict)


class EmptyReader(BaseReader):
    """
    This class does nothing.
    """

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name,
        cls_streamer_info,
        all_streamer_info,
        item_path,
    ):
        return None

    @classmethod
    def get_reader_instance(cls, tree_config: dict):
        if tree_config["reader"] != ReaderType.Empty:
            return None

        return bcpp.EmptyReader(tree_config["name"])

    @classmethod
    def reconstruct_array(cls, raw_data, tree_config):
        if tree_config["reader"] != ReaderType.Empty:
            return None

        return np.empty(shape=(0,))


class Bes3TObjArrayReader(BaseReader):
    bes3_branch2types = {
        "/Event:TMcEvent/m_mdcMcHitCol": "TMdcMc",
        "/Event:TMcEvent/m_cgemMcHitCol": "TCgemMc",
        "/Event:TMcEvent/m_emcMcHitCol": "TEmcMc",
        "/Event:TMcEvent/m_tofMcHitCol": "TTofMc",
        "/Event:TMcEvent/m_mucMcHitCol": "TMucMc",
        "/Event:TMcEvent/m_mcParticleCol": "TMcParticle",
        "/Event:TDigiEvent/m_mdcDigiCol": "TMdcDigi",
        "/Event:TDigiEvent/m_cgemDigiCol": "TCgemDigi",
        "/Event:TDigiEvent/m_emcDigiCol": "TEmcDigi",
        "/Event:TDigiEvent/m_tofDigiCol": "TTofDigi",
        "/Event:TDigiEvent/m_mucDigiCol": "TMucDigi",
        "/Event:TDigiEvent/m_lumiDigiCol": "TLumiDigi",
        "/Event:TDstEvent/m_mdcTrackCol": "TMdcTrack",
        "/Event:TDstEvent/m_emcTrackCol": "TEmcTrack",
        "/Event:TDstEvent/m_tofTrackCol": "TTofTrack",
        "/Event:TDstEvent/m_mucTrackCol": "TMucTrack",
        "/Event:TDstEvent/m_mdcDedxCol": "TMdcDedx",
        "/Event:TDstEvent/m_extTrackCol": "TExtTrack",
        "/Event:TDstEvent/m_mdcKalTrackCol": "TMdcKalTrack",
        "/Event:TRecEvent/m_recMdcTrackCol": "TRecMdcTrack",
        "/Event:TRecEvent/m_recMdcHitCol": "TRecMdcHit",
        "/Event:TRecEvent/m_recEmcHitCol": "TRecEmcHit",
        "/Event:TRecEvent/m_recEmcClusterCol": "TRecEmcCluster",
        "/Event:TRecEvent/m_recEmcShowerCol": "TRecEmcShower",
        "/Event:TRecEvent/m_recTofTrackCol": "TRecTofTrack",
        "/Event:TRecEvent/m_recMucTrackCol": "TRecMucTrack",
        "/Event:TRecEvent/m_recMdcDedxCol": "TRecMdcDedx",
        "/Event:TRecEvent/m_recMdcDedxHitCol": "TRecMdcDedxHit",
        "/Event:TRecEvent/m_recExtTrackCol": "TRecExtTrack",
        "/Event:TRecEvent/m_recMdcKalTrackCol": "TRecMdcKalTrack",
        "/Event:TRecEvent/m_recMdcKalHelixSegCol": "TRecMdcKalHelixSeg",
        "/Event:TRecEvent/m_recEvTimeCol": "TRecEvTime",
        "/Event:TRecEvent/m_recZddChannelCol": "TRecZddChannel",
        "/Event:TEvtRecObject/m_evtRecTrackCol": "TEvtRecTrack",
        "/Event:TEvtRecObject/m_evtRecVeeVertexCol": "TEvtRecVeeVertex",
        "/Event:TEvtRecObject/m_evtRecPi0Col": "TEvtRecPi0",
        "/Event:TEvtRecObject/m_evtRecEtaToGGCol": "TEvtRecEtaToGG",
        "/Event:TEvtRecObject/m_evtRecDTagCol": "TEvtRecDTag",
        "/Event:THltEvent/m_hltRawCol": "THltRaw",
        "/Event:EventNavigator/m_mcMdcMcHits": "map<int,int>",
        "/Event:EventNavigator/m_mcMdcTracks": "map<int,int>",
        "/Event:EventNavigator/m_mcEmcMcHits": "map<int,int>",
        "/Event:EventNavigator/m_mcEmcRecShowers": "map<int,int>",
    }

    @classmethod
    def gen_tree_config(
        cls,
        top_type_name: str,
        cls_streamer_info: dict,
        all_streamer_info: dict,
        item_path: str = "",
    ):
        if top_type_name != "TObjArray":
            return None

        obj_typename = Bes3TObjArrayReader.bes3_branch2types.get(
            item_path.replace(".TObjArray*", "")
        )
        if obj_typename is None:
            return None

        if obj_typename not in all_streamer_info:
            return {
                "reader": "MyTObjArrayReader",
                "name": cls_streamer_info["fName"],
                "element_reader": {
                    "reader": ReaderType.Empty,
                    "name": obj_typename,
                },
            }

        sub_reader_config = []
        for s in all_streamer_info[obj_typename]:
            sub_reader_config.append(
                gen_tree_config(s, all_streamer_info, item_path + f".{obj_typename}")
            )

        return {
            "reader": "MyTObjArrayReader",
            "name": cls_streamer_info["fName"],
            "element_reader": {
                "reader": ReaderType.ObjectHeader,
                "name": obj_typename,
                "sub_readers": sub_reader_config,
            },
        }

    @staticmethod
    def get_reader_instance(reader_config: dict):
        if reader_config["reader"] != "MyTObjArrayReader":
            return None

        element_reader_config = reader_config["element_reader"]
        element_reader = get_reader_instance(element_reader_config)

        return bcpp.Bes3TObjArrayReader(reader_config["name"], element_reader)

    @staticmethod
    def reconstruct_array(raw_data, reader_config: dict):
        if reader_config["reader"] != "MyTObjArrayReader":
            return None

        counts, element_raw_data = raw_data
        element_reader_config = reader_config["element_reader"]
        element_data = reconstruct_array(
            element_raw_data,
            element_reader_config,
        )

        return ak.unflatten(element_data, counts)


readers |= {
    CTypeReader,
    STLSequenceReader,
    STLMapReader,
    STLStringReader,
    TArrayReader,
    TStringReader,
    TObjectReader,
    CArrayReader,
    BaseObjectReader,
    ObjectHeaderReader,
    EmptyReader,
    Bes3TObjArrayReader,
}


##########################################################################################
#                                     Array Preprocess
##########################################################################################
def get_symetric_matrix_idx(
    i: Union[int, ak.Array, np.ndarray], j: Union[int, ak.Array, np.ndarray], ndim: int
) -> int:
    """
    Returns the index of the similarity matrix given the row and column indices.

    The matrix is assumed to be symmetric-like. (i, j) -> index relationship is:

    |     | i=0 | i=1 | i=2 |
    | :-: | :-: | :-: | :-: |
    | j=0 |  0  |     |     |
    | j=1 |  1  |  2  |     |
    | j=2 |  3  |  4  |  5  |

    Parameters:
        i (Union[int, ak.Array, np.ndarray]): The row index or array of row indices.
        j (Union[int, ak.Array, np.ndarray]): The column index or array of column indices.
        ndim (int): The dimension of the similarity matrix.

    Returns:
        The index or array of indices corresponding to the given row and column indices.

    Raises:
        ValueError: If the row and column indices are not of the same type, or if one of them is not an integer.
        ValueError: If the row or column indices are greater than or equal to the dimension of the similarity matrix.
        ValueError: If the row or column indices are negative.
    """
    # Check type
    return_type: Literal["ak", "np"] = "ak"
    if type(i) != type(j):
        if isinstance(i, int):
            return_type = "np" if isinstance(j, np.ndarray) else "ak"
            i = ak.ones_like(j) * i
        elif isinstance(j, int):
            return_type = "np" if isinstance(i, np.ndarray) else "ak"
            j = ak.ones_like(i) * j
        else:
            raise ValueError(
                "i and j should be the same type, or one of them should be an integer."
            )
    else:
        return_type = "np" if isinstance(i, np.ndarray) else "ak"

    i, j = ak.sort([i, j], axis=0)
    res = j * (j + 1) // 2 + i

    # Check dimension
    if ak.any([i >= ndim, j >= ndim]):
        raise ValueError(
            "Indices i and j should be less than the dimension of the similarity matrix."
        )
    if ak.any([i < 0, j < 0]):
        raise ValueError("Indices i and j should be non-negative.")

    if return_type == "np" and isinstance(res, ak.Array):
        res = res.to_numpy()

    return res


def expand_zipped_symetric_matrix(
    arr: Union[ak.Array, np.ndarray],
) -> Union[ak.Array, np.ndarray]:
    """
    Recover a flattened simplified symmetric matrix represented as a 1D array back to a 2D matrix.
    This function assumes the last dimension of the input array is the flattened symmetric matrix,
    and will transform array

    ```
    [[a11, a12, a22, a13, a23, a33],
     [b11, b12, b22, b13, b23, b33]]
    ```

    to

    ```
    [[[a11, a12, a13],
      [a12, a22, a23],
      [a13, a23, a33]],

     [[b11, b12, b13],
      [b12, b22, b23],
      [b13, b23, b33]]]
    ```

    Parameters:
        arr (Union[ak.Array, np.ndarray]): The input array representing the flattened simplified symmetric matrix.

    Returns:
        The reshaped symmetric matrix as a 2D array.

    Raises:
        ValueError: If the input array does not have a symmetric shape.
    """

    # Get the number of elements in the symmetric matrix
    if isinstance(arr, ak.Array):
        type_strs = [i.strip() for i in arr.typestr.split("*")[:-1]]
        n_err_elements = int(type_strs[-1])
        raw_shape = _extract_index(arr.layout)[:-1]
        flat_arr = _flat_to_numpy(arr).flatten().reshape(-1, n_err_elements)
    else:
        n_err_elements = arr.shape[-1]
        raw_shape = arr.shape[:-1]
        flat_arr = arr.reshape(-1, n_err_elements)

    ndim_err = (np.sqrt(1 + 8 * n_err_elements) - 1) / 2
    if not ndim_err.is_integer():
        raise ValueError("The array does not have a symmetric shape.")
    ndim_err = int(ndim_err)

    # Preapre output array
    n_raw_len = len(flat_arr.flatten())
    n_out_len = n_raw_len // n_err_elements * (ndim_err**2)
    raw_out = np.zeros(n_out_len, dtype=flat_arr.dtype).reshape(-1, ndim_err, ndim_err)

    # Fill error matrix
    for i in range(ndim_err):
        for j in range(ndim_err):
            idx = get_symetric_matrix_idx(i, j, ndim_err)
            raw_out[:, i, j] = flat_arr[:, idx]

    # Reshape the output array to match the original shape
    if isinstance(arr, ak.Array):
        res = _recover_shape(ak.Array(raw_out), raw_shape)
    else:
        res = raw_out.reshape(*raw_shape, ndim_err, ndim_err)

    return res


def expand_subbranch_symetric_matrix(
    sub_br_arr: ak.Array, matrix_fields: Union[str, set[str]]
) -> ak.Array:
    """
    Recover simplified symmetric matrix back to 2D matrix from specified fields of a branch array.

    Parameters:
        sub_br_arr: Subbranch array that need to be recovered.
        matrix_fields: Name of list of names of fields to be recovered.

    Returns:
        An array with recovered fields.
    """
    if isinstance(matrix_fields, str):
        matrix_fields = {matrix_fields}
    matrix_fields = set(matrix_fields)

    raw_shape = _extract_index(sub_br_arr.layout)

    res_dict = {}
    for field_name in sub_br_arr.fields:
        flat_sub_arr = sub_br_arr[field_name]
        for _ in range(len(raw_shape)):
            flat_sub_arr = ak.flatten(flat_sub_arr)

        if field_name in matrix_fields:
            res_dict[field_name] = expand_zipped_symetric_matrix(flat_sub_arr)
        else:
            res_dict[field_name] = flat_sub_arr

    res_arr = _recover_shape(ak.Array(res_dict), raw_shape)
    return res_arr


#############################################
# TDigiEvent
#############################################
def process_digi_subbranch(org_arr: ak.Array) -> ak.Array:
    """
    Processes the `TRawData` subbranch of the input awkward array and returns a new array with the subbranch fields
    merged into the top level.

    Parameters:
        org_arr (ak.Array): The input awkward array containing the `TRawData` subbranch.

    Returns:
        A new awkward array with the fields of `TRawData` merged into the top level.

    Raises:
        AssertionError: If `TRawData` is not found in the input array fields.
    """
    assert "TRawData" in org_arr.fields, "TRawData not found in the input array"

    fields = {}
    for field_name in org_arr.fields:
        if field_name == "TRawData":
            for raw_field_name in org_arr[field_name].fields:
                fields[raw_field_name] = org_arr[field_name][raw_field_name]
        else:
            fields[field_name] = org_arr[field_name]

    return ak.Array(fields)


#############################################
# TEvtRecObject
#############################################
def process_evtrec_m_Evtx(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_Evtx")


def process_evtrec_m_evtRecVeeVertexCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_Ew")


#############################################
# TDstEvent
#############################################
def process_dst_m_mdcTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_dst_m_emcTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_dst_m_extTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(
        org_arr,
        {
            "myEmcErrorMatrix",
            "myMucErrorMatrix",
            "myTof1ErrorMatrix",
            "myTof2ErrorMatrix",
        },
    )


def process_dst_m_mdcKalTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(
        org_arr,
        {
            "m_zerror",
            "m_zerror_e",
            "m_zerror_mu",
            "m_zerror_k",
            "m_zerror_p",
            "m_ferror",
            "m_ferror_e",
            "m_ferror_mu",
            "m_ferror_k",
            "m_ferror_p",
        },
    )


#############################################
# TRecEvent
#############################################
def process_rec_m_recMdcTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_rec_m_recEmcShowerCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_rec_m_recMdcKalTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_terror")


#############################################
# Main function
#############################################
def preprocess_subbranch(full_branch_path: str, org_arr: ak.Array) -> ak.Array:
    full_branch_path = full_branch_path.replace("/Event:", "")
    evt_name, subbranch_name = full_branch_path.split("/")

    if evt_name == "TDigiEvent" and subbranch_name != "m_fromMc":
        return process_digi_subbranch(org_arr)

    if evt_name == "TEvtRecObject":
        if subbranch_name == "m_Evtx":
            return process_evtrec_m_Evtx(org_arr)
        if subbranch_name == "m_evtRecVeeVertexCol":
            return process_evtrec_m_evtRecVeeVertexCol(org_arr)

    if evt_name == "TDstEvent":
        if subbranch_name == "m_mdcTrackCol":
            return process_dst_m_mdcTrackCol(org_arr)
        if subbranch_name == "m_emcTrackCol":
            return process_dst_m_emcTrackCol(org_arr)
        if subbranch_name == "m_extTrackCol":
            return process_dst_m_extTrackCol(org_arr)
        if subbranch_name == "m_mdcKalTrackCol":
            return process_dst_m_mdcKalTrackCol(org_arr)

    if evt_name == "TRecEvent":
        if subbranch_name == "m_recMdcTrackCol":
            return process_rec_m_recMdcTrackCol(org_arr)
        if subbranch_name == "m_recEmcShowerCol":
            return process_rec_m_recEmcShowerCol(org_arr)
        if subbranch_name == "m_recMdcKalTrackCol":
            return process_rec_m_recMdcKalTrackCol(org_arr)

    # Default return
    return org_arr


class Bes3Interpretation(uproot.interpretation.Interpretation):
    """
    Custom interpretation for Bes3 data.
    """

    target_branches: set[str] = set(Bes3TObjArrayReader.bes3_branch2types.keys())

    def __init__(
        self,
        branch: uproot.behaviors.TBranch.TBranch,
        context: dict,
        simplify: bool,
    ):
        """
        Args:
            branch (:doc:`uproot.behaviors.TBranch.TBranch`): The ``TBranch`` to
                interpret as an array.
            context (dict): Auxiliary data used in deserialization.
            simplify (bool): If True, call
                :ref:`uproot.interpretation.objects.AsObjects.simplify` on any
                :doc:`uproot.interpretation.objects.AsObjects` to try to get a
                more efficient interpretation.

        Accept arguments from `uproot.interpretation.identify.interpretation_of`.
        """
        self._branch = branch
        self._context = context
        self._simplify = simplify

        # simplify streamer information
        self.all_streamer_info: dict[str, list[dict]] = {}
        for k, v in branch.file.streamers.items():
            cur_infos = [i.all_members for i in next(iter(v.values())).member("fElements")]
            self.all_streamer_info[k] = cur_infos

    @classmethod
    def match_branch(
        cls,
        branch: uproot.behaviors.TBranch.TBranch,
        context: dict,
        simplify: bool,
    ) -> bool:
        """
        Args:
            branch (:doc:`uproot.behaviors.TBranch.TBranch`): The ``TBranch`` to
                interpret as an array.
            context (dict): Auxiliary data used in deserialization.
            simplify (bool): If True, call
                :ref:`uproot.interpretation.objects.AsObjects.simplify` on any
                :doc:`uproot.interpretation.objects.AsObjects` to try to get a
                more efficient interpretation.

        Accept arguments from `uproot.interpretation.identify.interpretation_of`,
        determine whether this interpretation can be applied to the given branch.
        """
        full_path = regularize_object_path(branch.object_path)
        return full_path in cls.target_branches

    @property
    def typename(self) -> str:
        """
        The name of the type of the interpretation.
        """
        return self._branch.streamer.typename

    @property
    def cache_key(self) -> str:
        """
        The cache key of the interpretation.
        """
        return id(self)

    def __repr__(self) -> str:
        """
        The string representation of the interpretation.
        """
        return f"AsBes3Custom({self.typename})"

    def final_array(
        self,
        basket_arrays,
        entry_start,
        entry_stop,
        entry_offsets,
        library,
        branch,
        options,
    ):
        """
        Concatenate the arrays from the baskets and return the final array.
        """

        awkward = uproot.extras.awkward()

        basket_entry_starts = np.array(entry_offsets[:-1])
        basket_entry_stops = np.array(entry_offsets[1:])

        basket_start_idx = np.where(basket_entry_starts <= entry_start)[0].max()
        basket_end_idx = np.where(basket_entry_stops >= entry_stop)[0].min()

        arr_to_concat = [basket_arrays[i] for i in range(basket_start_idx, basket_end_idx + 1)]
        tot_array = awkward.concatenate(arr_to_concat)

        relative_entry_start = entry_start - basket_entry_starts[basket_start_idx]
        relative_entry_stop = entry_stop - basket_entry_starts[basket_start_idx]

        return tot_array[relative_entry_start:relative_entry_stop]

    def basket_array(
        self,
        data,
        byte_offsets,
        basket,
        branch,
        context,
        cursor_offset,
        library,
        interp_options,
    ):
        assert library.name == "ak", "Only awkward arrays are supported"

        full_branch_path = regularize_object_path(branch.object_path)

        # generate reader config
        tree_config = gen_tree_config_from_type_name(
            branch.streamer.typename, self.all_streamer_info, full_branch_path
        )

        # get reader
        reader = get_reader_instance(tree_config)

        # do read
        raw_data = bcpp.read_data(data, byte_offsets, reader)

        # recover raw data
        raw_ak_arr = reconstruct_array(raw_data, tree_config)

        # preprocess awkward array and return
        return preprocess_subbranch(full_branch_path, raw_ak_arr)


##########################################################################################
#                                       Wrappers
##########################################################################################
_is_TBranchElement_branches_wrapped = False
_is_uproot_interpretation_of_wrapped = False

_uproot_interpretation_of = uproot.interpretation.identify.interpretation_of


def bes_interpretation_of(
    branch: uproot.behaviors.TBranch.TBranch, context: dict, simplify: bool = True
) -> uproot.interpretation.Interpretation:
    if not hasattr(branch, "parent"):
        return _uproot_interpretation_of(branch, context, simplify)

    if Bes3Interpretation.match_branch(branch, context, simplify):
        return Bes3Interpretation(branch, context, simplify)

    return _uproot_interpretation_of(branch, context, simplify)


def wrap_uproot_interpretation():
    global _is_uproot_interpretation_of_wrapped
    if not _is_uproot_interpretation_of_wrapped:
        _is_uproot_interpretation_of_wrapped = True
        uproot.interpretation.identify.interpretation_of = bes_interpretation_of


def wrap_uproot_TBranchElement_branches():
    def branches(self):

        if self.name not in {
            "TEvtHeader",
            "TMcEvent",
            "TDigiEvent",
            "TDstEvent",
            "TRecEvent",
            "TEvtRecObject",
            "THltEvent",
        }:
            return self.member("fBranches")
        else:

            res = []
            for br in self.member("fBranches"):
                if br.name == "TObject":
                    continue

                full_path = regularize_object_path(br.object_path)
                if full_path not in Bes3TObjArrayReader.bes3_branch2types:
                    res.append(br)
                    continue

                class_name = Bes3TObjArrayReader.bes3_branch2types[full_path]
                if class_name in self.file.streamers:
                    res.append(br)
                else:
                    continue
            return res

    global _is_TBranchElement_branches_wrapped
    if not _is_TBranchElement_branches_wrapped:
        _is_TBranchElement_branches_wrapped = True
        uproot.models.TBranch.Model_TBranchElement.branches = property(branches)
        for v in uproot.models.TBranch.Model_TBranchElement.known_versions.values():
            v.branches = property(branches)


def wrap_uproot():
    """
    Wraps the uproot functions to use the BES interpretation.
    """
    wrap_uproot_interpretation()
    wrap_uproot_TBranchElement_branches()



import vector

vector.register_awkward()


from .helix import (
    HelixObject,
    dr_phi0_to_x,
    dr_phi0_to_y,
    helix_awk,
    helix_obj,
    kappa_to_charge,
    kappa_to_pt,
    kappa_to_radius,
    phi0_to_phi,
)
from .old_helix import compute_helix, parse_helix, regularize_helix



import os
from glob import glob
from pathlib import Path
from typing import Union

cur_dir = Path(__file__).parent
geom_dir = cur_dir / "detectors/geometry"

src_cache_list = [
    (geom_dir / "mdc_geom.npz", geom_dir / "__pycache__/mdc.*.nb[ci]"),
    (geom_dir / "emc_geom.npz", geom_dir / "__pycache__/emc.*.nb[ci]"),
]


def cache_auto_clear(
    sources: Union[Union[Path, str], list[Union[Path, str]]],
    caches: Union[Union[Path, str], list[Union[Path, str]]],
    silent: bool = True,
    force: bool = False,
) -> list[str]:
    """
    Automatically clear the Numba cache if the source files have changed.

    Due to limitation of numba cache, modification of external files like
    geometry data will not trigger the cache clearing. This function is
    designed to trigger the cache clearing when those external files are
    modified.

    This function compares the modification times of the source files with the
    modification times of the cached files. If any source file is newer than
    the cached files, the numba cache in cache directory will be cleared.

    !!! note
        There is no need to trace `*.py` files, as numba will automatically
        clear the cache when the source code is modified.

    Args:
        sources: A glob pattern or a list of glob patterns to match the source files.
        caches: A glob pattern or a list of glob patterns to match the cached files.
        silent: If True, suppresses the output messages.
        force: If True, forces the cache clearing even if the source files are not newer.

    Returns:
        list[str]: A list of cache files that were removed.
    """
    # Transform sources and caches to lists of strings
    sources: list[str] = (
        glob(str(sources))
        if isinstance(sources, (Path, str))
        else sum([glob(str(s)) for s in sources], [])
    )

    caches: list[str] = (
        glob(str(caches))
        if isinstance(caches, (Path, str))
        else sum([glob(str(c)) for c in caches], [])
    )

    # Check if sources and caches are empty
    if not sources:
        raise ValueError("No source files found.")

    if not caches:
        return []

    # Compare modification times
    removed_caches = []
    failed_removed_caches = []

    src_latest_mtime = max([os.path.getmtime(s) for s in sources if os.path.isfile(s)])
    cache_earliest_mtime = min([os.path.getmtime(c) for c in caches if os.path.isfile(c)])

    if src_latest_mtime > cache_earliest_mtime or force:
        # Remove the cache files
        for c in caches:
            try:
                os.remove(c)
                removed_caches.append(c)
            except FileNotFoundError:
                pass
            except Exception:
                failed_removed_caches.append(c)

    # Check failure of removing
    if failed_removed_caches:
        error_msg = "Cannot remove cache files:\n"
        for c in failed_removed_caches:
            error_msg += f" - {c}\n"
        error_msg += "This may cause wrong results from corresponding functions. "
        error_msg += "Please check their permissions or remove them manually."
        raise ImportError(error_msg)

    # Print messages if not silent
    if not silent:
        if removed_caches:
            removed_caches_str = "\n - ".join(removed_caches)
            print(f"Removed cache files: {removed_caches_str}")

    # Return the list of removed caches
    return removed_caches


def check_numba_cache():
    """
    Check the cache files and remove them if the source files are newer.
    This function should be called when the module is imported.

    When environment variable `PYBES3_NUMBA_CACHE_SILENT` is set to 1,
    the function will print out removal messages.
    """
    silent = os.getenv("PYBES3_NUMBA_CACHE_SILENT", "0") == "0"
    for src, cache in src_cache_list:
        cache_auto_clear(
            sources=src,
            caches=cache,
            silent=silent,
            force=False,
        )


def clear_numba_cache():
    """
    Clear the cache files regardless of the source files.

    When environment variable `PYBES3_NUMBA_CACHE_SILENT` is set to 1,
    the function will print out removal messages.
    """
    silent = os.getenv("PYBES3_NUMBA_CACHE_SILENT", "0") == "0"
    for src, cache in src_cache_list:
        cache_auto_clear(
            sources=src,
            caches=cache,
            silent=silent,
            force=True,
        )



from pathlib import Path

import awkward as ak
import numpy as np

import pybes3 as p3
from pybes3 import detectors as det
from pybes3.detectors import digi_id

data_dir = Path(__file__).parent.parent / "data"
rtraw_event = p3.open(data_dir / "test_full_mc_evt_1.rtraw")["Event/TDigiEvent"].arrays()


def test_mdc_digi_id():
    mdc_id_ak: ak.Array = rtraw_event["m_mdcDigiCol"]["m_intId"]

    # Test awkward
    ak_wire, ak_layer, ak_is_stereo = (
        digi_id.mdc_id_to_wire(mdc_id_ak),
        digi_id.mdc_id_to_layer(mdc_id_ak),
        digi_id.mdc_id_to_is_stereo(mdc_id_ak),
    )
    assert ak.all(det.get_mdc_digi_id(ak_wire, ak_layer, ak_is_stereo) == mdc_id_ak)
    assert ak.all(digi_id.check_mdc_id(mdc_id_ak))

    # Test numpy
    mdc_id_np = ak.flatten(mdc_id_ak).to_numpy()
    np_wire, np_layer, np_is_stereo = (
        digi_id.mdc_id_to_wire(mdc_id_np),
        digi_id.mdc_id_to_layer(mdc_id_np),
        digi_id.mdc_id_to_is_stereo(mdc_id_np),
    )
    assert np.all(det.get_mdc_digi_id(np_wire, np_layer, np_is_stereo) == mdc_id_np)
    assert np.all(digi_id.check_mdc_id(mdc_id_np))

    # Test uint32
    tmp_id = mdc_id_np[0]
    tmp_wire, tmp_layer, tmp_is_stereo = (
        digi_id.mdc_id_to_wire(tmp_id),
        digi_id.mdc_id_to_layer(tmp_id),
        digi_id.mdc_id_to_is_stereo(tmp_id),
    )
    assert tmp_wire == np_wire[0]
    assert tmp_layer == np_layer[0]
    assert tmp_is_stereo == np_is_stereo[0]
    assert det.get_mdc_digi_id(tmp_wire, tmp_layer, tmp_is_stereo) == tmp_id
    assert digi_id.check_mdc_id(tmp_id)

    # Test python int
    tmp_id = int(mdc_id_np[0])
    tmp_wire, tmp_layer, tmp_is_stereo = (
        digi_id.mdc_id_to_wire(tmp_id),
        digi_id.mdc_id_to_layer(tmp_id),
        digi_id.mdc_id_to_is_stereo(tmp_id),
    )
    assert tmp_wire == np_wire[0]
    assert tmp_layer == np_layer[0]
    assert tmp_is_stereo == np_is_stereo[0]
    assert det.get_mdc_digi_id(tmp_wire, tmp_layer, tmp_is_stereo) == tmp_id
    assert digi_id.check_mdc_id(tmp_id)


def test_emc_digi_id():
    emc_id_ak: ak.Array = rtraw_event["m_emcDigiCol"]["m_intId"]

    # Test awkward
    ak_module, ak_theta, ak_phi = (
        digi_id.emc_id_to_module(emc_id_ak),
        digi_id.emc_id_to_theta(emc_id_ak),
        digi_id.emc_id_to_phi(emc_id_ak),
    )
    assert ak.all(det.get_emc_digi_id(ak_module, ak_theta, ak_phi) == emc_id_ak)
    assert ak.all(digi_id.check_emc_id(emc_id_ak))

    # Test numpy
    emc_id_np = ak.flatten(emc_id_ak).to_numpy()
    np_module, np_theta, np_phi = (
        digi_id.emc_id_to_module(emc_id_np),
        digi_id.emc_id_to_theta(emc_id_np),
        digi_id.emc_id_to_phi(emc_id_np),
    )
    assert np.all(det.get_emc_digi_id(np_module, np_theta, np_phi) == emc_id_np)
    assert np.all(digi_id.check_emc_id(emc_id_np))

    # Test uint32
    tmp_id = emc_id_np[0]
    tmp_module, tmp_theta, tmp_phi = (
        digi_id.emc_id_to_module(tmp_id),
        digi_id.emc_id_to_theta(tmp_id),
        digi_id.emc_id_to_phi(tmp_id),
    )
    assert tmp_module == np_module[0]
    assert tmp_theta == np_theta[0]
    assert tmp_phi == np_phi[0]
    assert det.get_emc_digi_id(tmp_module, tmp_theta, tmp_phi) == tmp_id
    assert digi_id.check_emc_id(tmp_id)

    # Test python int
    tmp_id = int(emc_id_np[0])
    tmp_module, tmp_theta, tmp_phi = (
        digi_id.emc_id_to_module(tmp_id),
        digi_id.emc_id_to_theta(tmp_id),
        digi_id.emc_id_to_phi(tmp_id),
    )
    assert tmp_module == np_module[0]
    assert tmp_theta == np_theta[0]
    assert tmp_phi == np_phi[0]
    assert det.get_emc_digi_id(tmp_module, tmp_theta, tmp_phi) == tmp_id
    assert digi_id.check_emc_id(tmp_id)


def test_tof_digi_id():
    tof_id_ak: ak.Array = rtraw_event["m_tofDigiCol"]["m_intId"]

    # Test awkward
    ak_part, ak_layer_or_module, ak_phi_or_strip, ak_end = (
        digi_id.tof_id_to_part(tof_id_ak),
        digi_id.tof_id_to_layer_or_module(tof_id_ak),
        digi_id.tof_id_to_phi_or_strip(tof_id_ak),
        digi_id.tof_id_to_end(tof_id_ak),
    )
    assert ak.all(
        det.get_tof_digi_id(ak_part, ak_layer_or_module, ak_phi_or_strip, ak_end) == tof_id_ak
    )
    assert ak.all(digi_id.check_tof_id(tof_id_ak))

    # Test numpy
    tof_id_np = ak.flatten(tof_id_ak).to_numpy()
    np_part, np_layer_or_module, np_phi_or_strip, np_end = (
        digi_id.tof_id_to_part(tof_id_np),
        digi_id.tof_id_to_layer_or_module(tof_id_np),
        digi_id.tof_id_to_phi_or_strip(tof_id_np),
        digi_id.tof_id_to_end(tof_id_np),
    )
    assert np.all(
        det.get_tof_digi_id(np_part, np_layer_or_module, np_phi_or_strip, np_end) == tof_id_np
    )
    assert np.all(digi_id.check_tof_id(tof_id_np))

    # Test uint32
    tmp_id = tof_id_np[0]
    tmp_part, tmp_layer_or_module, tmp_phi_or_strip, tmp_end = (
        digi_id.tof_id_to_part(tmp_id),
        digi_id.tof_id_to_layer_or_module(tmp_id),
        digi_id.tof_id_to_phi_or_strip(tmp_id),
        digi_id.tof_id_to_end(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_layer_or_module == np_layer_or_module[0]
    assert tmp_phi_or_strip == np_phi_or_strip[0]
    assert tmp_end == np_end[0]
    assert (
        det.get_tof_digi_id(tmp_part, tmp_layer_or_module, tmp_phi_or_strip, tmp_end) == tmp_id
    )
    assert digi_id.check_tof_id(tmp_id)

    # Test python int
    tmp_id = int(tof_id_np[0])
    tmp_part, tmp_layer_or_module, tmp_phi_or_strip, tmp_end = (
        digi_id.tof_id_to_part(tmp_id),
        digi_id.tof_id_to_layer_or_module(tmp_id),
        digi_id.tof_id_to_phi_or_strip(tmp_id),
        digi_id.tof_id_to_end(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_layer_or_module == np_layer_or_module[0]
    assert tmp_phi_or_strip == np_phi_or_strip[0]
    assert tmp_end == np_end[0]
    assert (
        det.get_tof_digi_id(tmp_part, tmp_layer_or_module, tmp_phi_or_strip, tmp_end) == tmp_id
    )
    assert digi_id.check_tof_id(tmp_id)


def test_muc_digi_id():
    muc_id_ak: ak.Array = rtraw_event["m_mucDigiCol"]["m_intId"]

    # Test awkward
    ak_part, ak_segment, ak_layer, ak_channel, ak_gap, ak_strip = (
        digi_id.muc_id_to_part(muc_id_ak),
        digi_id.muc_id_to_segment(muc_id_ak),
        digi_id.muc_id_to_layer(muc_id_ak),
        digi_id.muc_id_to_channel(muc_id_ak),
        digi_id.muc_id_to_gap(muc_id_ak),
        digi_id.muc_id_to_strip(muc_id_ak),
    )
    assert ak.all(det.get_muc_digi_id(ak_part, ak_segment, ak_layer, ak_channel) == muc_id_ak)
    assert ak.all(digi_id.check_muc_id(muc_id_ak))
    assert ak.all(ak_layer == ak_gap)
    assert ak.all(ak_channel == ak_strip)

    # Test numpy
    muc_id_np = ak.flatten(muc_id_ak).to_numpy()
    np_part, np_segment, np_layer, np_channel, np_gap, np_strip = (
        digi_id.muc_id_to_part(muc_id_np),
        digi_id.muc_id_to_segment(muc_id_np),
        digi_id.muc_id_to_layer(muc_id_np),
        digi_id.muc_id_to_channel(muc_id_np),
        digi_id.muc_id_to_gap(muc_id_np),
        digi_id.muc_id_to_strip(muc_id_np),
    )
    assert np.all(det.get_muc_digi_id(np_part, np_segment, np_layer, np_channel) == muc_id_np)
    assert np.all(digi_id.check_muc_id(muc_id_np))
    assert np.all(np_layer == np_gap)
    assert np.all(np_channel == np_strip)

    # Test uint32
    tmp_id = muc_id_np[0]
    tmp_part, tmp_segment, tmp_layer, tmp_channel, tmp_gap, tmp_strip = (
        digi_id.muc_id_to_part(tmp_id),
        digi_id.muc_id_to_segment(tmp_id),
        digi_id.muc_id_to_layer(tmp_id),
        digi_id.muc_id_to_channel(tmp_id),
        digi_id.muc_id_to_gap(tmp_id),
        digi_id.muc_id_to_strip(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_segment == np_segment[0]
    assert tmp_layer == np_layer[0]
    assert tmp_channel == np_channel[0]
    assert tmp_gap == np_gap[0]
    assert tmp_strip == np_strip[0]
    assert det.get_muc_digi_id(tmp_part, tmp_segment, tmp_layer, tmp_channel) == tmp_id
    assert digi_id.check_muc_id(tmp_id)

    # Test python int
    tmp_id = int(muc_id_np[0])
    tmp_part, tmp_segment, tmp_layer, tmp_channel, tmp_gap, tmp_strip = (
        digi_id.muc_id_to_part(tmp_id),
        digi_id.muc_id_to_segment(tmp_id),
        digi_id.muc_id_to_layer(tmp_id),
        digi_id.muc_id_to_channel(tmp_id),
        digi_id.muc_id_to_gap(tmp_id),
        digi_id.muc_id_to_strip(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_segment == np_segment[0]
    assert tmp_layer == np_layer[0]
    assert tmp_channel == np_channel[0]
    assert tmp_gap == np_gap[0]
    assert tmp_strip == np_strip[0]
    assert det.get_muc_digi_id(tmp_part, tmp_segment, tmp_layer, tmp_channel) == tmp_id
    assert digi_id.check_muc_id(tmp_id)


def test_cgem_digi_id():
    cgem_id_ak: ak.Array = p3.open(data_dir / "test_cgem.rtraw")[
        "Event/TDigiEvent/m_cgemDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_layer, ak_sheet, ak_strip, ak_is_x_strip = (
        digi_id.cgem_id_to_layer(cgem_id_ak),
        digi_id.cgem_id_to_sheet(cgem_id_ak),
        digi_id.cgem_id_to_strip(cgem_id_ak),
        digi_id.cgem_id_to_is_x_strip(cgem_id_ak),
    )
    assert ak.all(
        det.get_cgem_digi_id(ak_layer, ak_sheet, ak_strip, ak_is_x_strip) == cgem_id_ak
    )
    assert ak.all(digi_id.check_cgem_id(cgem_id_ak))

    # Test numpy
    cgem_id_np = ak.flatten(cgem_id_ak).to_numpy()
    np_layer, np_sheet, np_strip, np_is_x_strip = (
        digi_id.cgem_id_to_layer(cgem_id_np),
        digi_id.cgem_id_to_sheet(cgem_id_np),
        digi_id.cgem_id_to_strip(cgem_id_np),
        digi_id.cgem_id_to_is_x_strip(cgem_id_np),
    )
    assert np.all(
        det.get_cgem_digi_id(np_layer, np_sheet, np_strip, np_is_x_strip) == cgem_id_np
    )
    assert np.all(digi_id.check_cgem_id(cgem_id_np))

    # Test uint32
    tmp_id = cgem_id_np[0]
    tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip = (
        digi_id.cgem_id_to_layer(tmp_id),
        digi_id.cgem_id_to_sheet(tmp_id),
        digi_id.cgem_id_to_strip(tmp_id),
        digi_id.cgem_id_to_is_x_strip(tmp_id),
    )
    assert tmp_layer == np_layer[0]
    assert tmp_sheet == np_sheet[0]
    assert tmp_strip == np_strip[0]
    assert tmp_is_x_strip == np_is_x_strip[0]
    assert det.get_cgem_digi_id(tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip) == tmp_id
    assert digi_id.check_cgem_id(tmp_id)

    # Test python int
    tmp_id = int(cgem_id_np[0])
    tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip = (
        digi_id.cgem_id_to_layer(tmp_id),
        digi_id.cgem_id_to_sheet(tmp_id),
        digi_id.cgem_id_to_strip(tmp_id),
        digi_id.cgem_id_to_is_x_strip(tmp_id),
    )
    assert tmp_layer == np_layer[0]
    assert tmp_sheet == np_sheet[0]
    assert tmp_strip == np_strip[0]
    assert tmp_is_x_strip == np_is_x_strip[0]
    assert det.get_cgem_digi_id(tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip) == tmp_id
    assert digi_id.check_cgem_id(tmp_id)


if __name__ == "__main__":
    import pytest

    pytest.main([__file__, "-v", "-s"])



from typing import Any, Optional
import numpy as np
from numpy.typing import NDArray

class BaseReader:
    def __init__(self, name: str): ...
    def data(self) -> Any: ...

class Int8Reader(BaseReader):
    def data(self) -> NDArray[np.int8]: ...

class Int16Reader(BaseReader):
    def data(self) -> NDArray[np.int16]: ...

class Int32Reader(BaseReader):
    def data(self) -> NDArray[np.int32]: ...

class Int64Reader(BaseReader):
    def data(self) -> NDArray[np.int64]: ...

class UInt8Reader(BaseReader):
    def data(self) -> NDArray[np.uint8]: ...

class UInt16Reader(BaseReader):
    def data(self) -> NDArray[np.uint16]: ...

class UInt32Reader(BaseReader):
    def data(self) -> NDArray[np.uint32]: ...

class UInt64Reader(BaseReader):
    def data(self) -> NDArray[np.uint64]: ...

class FloatReader(BaseReader):
    def data(self) -> NDArray[np.float32]: ...

class DoubleReader(BaseReader):
    def data(self) -> NDArray[np.float64]: ...

class BoolReader(BaseReader):
    def data(self) -> NDArray[np.bool_]: ...

class STLSeqReader(BaseReader):
    def __init__(
        self,
        name: str,
        is_top: bool,
        element_reader: list[BaseReader],
    ): ...
    def data(self) -> tuple[NDArray[np.uint32], Any]: ...

class STLMapReader(BaseReader):
    def __init__(
        self,
        name: str,
        is_top: bool,
        key_reader: BaseReader,
        value_reader: BaseReader,
    ): ...
    def data(self) -> tuple[NDArray[np.uint32], Any, Any]: ...

class STLStringReader(BaseReader):
    def __init__(
        self,
        name: str,
        is_top: bool,
    ): ...
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.character]]: ...

class TArrayCReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.int8]]: ...

class TArraySReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.int16]]: ...

class TArrayIReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.int32]]: ...

class TArrayLReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.int64]]: ...

class TArrayFReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.float32]]: ...

class TArrayDReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.float64]]: ...

class TStringReader(BaseReader):
    def data(self) -> tuple[NDArray[np.uint32], NDArray[np.character]]: ...

class TObjectReader(BaseReader):
    def data(self) -> None: ...

class CArrayReader(BaseReader):
    def __init__(
        self,
        name: str,
        is_obj: bool,
        flat_size: int,
        element_reader: BaseReader,
    ): ...

class BaseObjectReader(BaseReader):
    def __init__(
        self,
        name: str,
        sub_readers: list[BaseReader],
    ): ...
    def data(self) -> list: ...

class ObjectHeaderReader(BaseReader):
    def __init__(
        self,
        name: str,
        sub_readers: list[BaseReader],
    ): ...
    def data(self) -> list: ...

class EmptyReader(BaseReader):
    def data(self) -> None: ...

class Bes3TObjArrayReader(BaseReader):
    def __init__(
        self,
        name: str,
        element_reader: BaseReader,
    ): ...
    def data(self) -> tuple[NDArray[np.uint32], Any]: ...

def read_data(
    data: NDArray[np.uint8],
    offsets: NDArray[np.uint32],
    reader: BaseReader,
) -> Any: ...
def read_bes_raw(
    data: NDArray[np.uint32],
    sub_detectors: Optional[list[str]] = None,
) -> dict: ...



from pathlib import Path

import awkward
import pytest
import uproot

import pybes3
import pybes3.besio as besio

besio.wrap_uproot()

data_dir = Path(__file__).parent / "data"


def test_uproot_branches():
    f_full = uproot.open(data_dir / "test_full_mc_evt_1.rtraw")
    assert len(f_full["Event/TMcEvent"].branches) == 5

    f_only_mc_particles = uproot.open(data_dir / "test_only_mc_particles.rtraw")
    assert len(f_only_mc_particles["Event/TMcEvent"].branches) == 1


def test_mc_full():
    f_rtraw = uproot.open(data_dir / "test_full_mc_evt_1.rtraw")
    arr = f_rtraw["Event"].arrays()
    assert len(arr) == 10


def test_mc_only_particles():
    f_rtraw = uproot.open(data_dir / "test_only_mc_particles.rtraw")
    arr = f_rtraw["Event"].arrays()
    assert len(arr) == 10


def test_dst():
    f_dst = uproot.open(data_dir / "test_full_mc_evt_1.dst")
    arr_dst = f_dst["Event"].arrays()
    assert len(arr_dst) == 10


def test_rec():
    f_rec = uproot.open(data_dir / "test_full_mc_evt_1.rec")
    arr_rec = f_rec["Event"].arrays()
    assert len(arr_rec) == 10


def test_cgem_rtraw():
    f_rtraw = uproot.open(data_dir / "test_cgem.rtraw")
    arr = f_rtraw["Event"].arrays()
    assert len(arr) == 200


def test_uproot_concatenate():
    arr_concat1 = uproot.concatenate(
        {
            data_dir / "test_full_mc_evt_1.rtraw": "Event",
            data_dir / "test_full_mc_evt_2.rtraw": "Event",
        }
    )
    assert len(arr_concat1) == 20

    arr_concat2 = uproot.concatenate(
        {
            data_dir / "test_full_mc_evt_1.rtraw": "Event/TMcEvent/m_mcParticleCol",
            data_dir / "test_full_mc_evt_2.rtraw": "Event/TMcEvent/m_mcParticleCol",
        }
    )
    assert len(arr_concat2) == 20


def test_bes_open():
    f = besio.open(data_dir / "test_full_mc_evt_1.rtraw")
    assert len(f["Event/TMcEvent"].branches) == 5

    f = besio.open(data_dir / "test_only_mc_particles.rtraw")
    assert len(f["Event/TMcEvent"].branches) == 1


def test_bes_concatenate():
    arr_concat1 = besio.concatenate(
        [data_dir / "test_full_mc_evt_1.rtraw", data_dir / "test_full_mc_evt_2.rtraw"],
        "Event/TMcEvent",
    )
    assert len(arr_concat1) == 20

    arr_concat2 = besio.concatenate(
        [data_dir / "test_full_mc_evt_1.rtraw", data_dir / "test_full_mc_evt_2.rtraw"],
        "Event/TMcEvent/m_mcParticleCol",
    )
    assert len(arr_concat2) == 20


def test_symetric_matrix_expansion():
    def test_symetric_matrix(arr):
        arr = awkward.flatten(arr)
        n_dim = int(arr.typestr.split("*")[-2].strip())

        # Check if the matrix is square
        assert n_dim == int(arr.typestr.split("*")[-3].strip())

        for i in range(n_dim):
            for j in range(i, n_dim):
                assert awkward.all(arr[:, i, j] == arr[:, j, i])

    f_dst = uproot.open(data_dir / "test_full_mc_evt_1.dst")
    arr_dst = f_dst["Event/TDstEvent"].arrays()

    f_rec = uproot.open(data_dir / "test_full_mc_evt_1.rec")
    arr_rec = f_rec["Event/TRecEvent"].arrays()

    for tmp_arr in [
        arr_dst.m_mdcTrackCol.m_err,
        arr_dst.m_emcTrackCol.m_err,
        arr_dst.m_extTrackCol.myEmcErrorMatrix,
        arr_dst.m_extTrackCol.myMucErrorMatrix,
        arr_dst.m_extTrackCol.myTof1ErrorMatrix,
        arr_dst.m_extTrackCol.myTof2ErrorMatrix,
        arr_rec.m_recMdcTrackCol.m_err,
        arr_rec.m_recEmcShowerCol.m_err,
        arr_rec.m_recMdcKalTrackCol.m_terror,
    ]:
        test_symetric_matrix(tmp_arr)


def test_digi_expand_TRawData():
    f_rec = uproot.open(data_dir / "test_full_mc_evt_1.rec")
    arr_digi = f_rec["Event/TDigiEvent"].arrays()
    for field in arr_digi.fields:
        if field == "m_fromMc":
            continue

        assert "TRawData" not in arr_digi[field].fields


@pytest.mark.skipif(
    not Path(
        "/bes3fs/offline/data/raw/round17/231117/run_0079017_All_file001_SFO-1.raw"
    ).exists(),
    reason="Test data is not available",
)
def test_raw():
    f_raw: pybes3.besio.RawBinaryReader = pybes3.open_raw(
        "/bes3fs/offline/data/raw/round17/231117/run_0079017_All_file001_SFO-1.raw"
    )

    n_mdc_digis = awkward.Array([1872, 2768, 1641, 2542, 3331, 2672, 2257, 2470, 3635, 3689])

    arr_full = f_raw.arrays(n_blocks=10)
    assert set(arr_full.fields) == {"evt_header", "mdc", "tof", "emc", "muc"}
    assert awkward.all(awkward.count(arr_full.mdc.id, axis=1) == n_mdc_digis)

    arr_mdc = f_raw.arrays(n_blocks=10, sub_detectors=["mdc"])
    assert set(arr_mdc.fields) == {"evt_header", "mdc"}
    assert awkward.all(awkward.count(arr_mdc.mdc.id, axis=1) == n_mdc_digis)

    arr_batch = f_raw.arrays(n_blocks=10, n_block_per_batch=5)
    assert set(arr_batch.fields) == {"evt_header", "mdc", "tof", "emc", "muc"}
    assert awkward.all(awkward.count(arr_batch.mdc.id, axis=1) == n_mdc_digis)

    arr_workers = f_raw.arrays(n_blocks=10, max_workers=4)
    assert set(arr_workers.fields) == {"evt_header", "mdc", "tof", "emc", "muc"}
    assert awkward.all(awkward.count(arr_workers.mdc.id, axis=1) == n_mdc_digis)


if __name__ == "__main__":
    import pytest

    pytest.main([__file__, "-v", "-s"])



import subprocess
import importlib.util

import pytest


@pytest.mark.skipif(
    importlib.util.find_spec("mkdocs") is None,
    reason="mkdocs is not installed, skipping build test",
)
def test_mkdocs_build():
    """Test that mkdocs build runs without errors."""
    try:
        result = subprocess.run(
            ["mkdocs", "build"], check=True, capture_output=True, text=True
        )
        assert result.returncode == 0, "mkdocs build failed"
    except subprocess.CalledProcessError as e:
        pytest.fail(f"mkdocs build failed with error: {e.stderr}")



from typing import Literal, Union

import awkward as ak
import numpy as np

from ..typing import IntLike
from . import digi_id
from .digi_id import (
    get_cgem_digi_id,
    get_emc_digi_id,
    get_mdc_digi_id,
    get_muc_digi_id,
    get_tof_digi_id,
)
from .geometry import (
    emc_barrel_h1,
    emc_barrel_h2,
    emc_barrel_h3,
    emc_barrel_l,
    emc_barrel_offset_1,
    emc_barrel_offset_2,
    emc_barrel_r,
    emc_gid_to_center_x,
    emc_gid_to_center_y,
    emc_gid_to_center_z,
    emc_gid_to_front_center_x,
    emc_gid_to_front_center_y,
    emc_gid_to_front_center_z,
    emc_gid_to_part,
    emc_gid_to_phi,
    emc_gid_to_point_x,
    emc_gid_to_point_y,
    emc_gid_to_point_z,
    emc_gid_to_theta,
    get_emc_crystal_position,
    get_emc_gid,
    get_mdc_gid,
    get_mdc_wire_position,
    mdc_gid_to_east_x,
    mdc_gid_to_east_y,
    mdc_gid_to_east_z,
    mdc_gid_to_is_stereo,
    mdc_gid_to_layer,
    mdc_gid_to_stereo,
    mdc_gid_to_superlayer,
    mdc_gid_to_west_x,
    mdc_gid_to_west_y,
    mdc_gid_to_west_z,
    mdc_gid_to_wire,
    mdc_gid_z_to_x,
    mdc_gid_z_to_y,
    mdc_layer_to_is_stereo,
    mdc_layer_to_superlayer,
)


###############################################################################
#                                     MDC                                     #
###############################################################################
def parse_mdc_gid(gid: IntLike, with_pos: bool = True) -> Union[IntLike, dict[str, IntLike]]:
    """
    Parse the gid of MDC wires. "gid" is the global ID of the wire, ranges from 0 to 6795.
    When `gid` is an `ak.Array`, the result is an `ak.Array`, otherwise it is a `dict`.

    Keys of the output:

    - `gid`: Global ID of the wire.
    - `layer`: Layer number.
    - `wire`: Local wire number.
    - `stereo`: Stereo type. 0 for axial, -1 for `phi_west < phi_east`, 1 for `phi_west > phi_east`.
    - `is_stereo`: Whether the wire is a stereo wire.
    - `superlayer`: Superlayer number.

    Optional keys of the output when `with_pos` is `True`:

    - `mid_x`: x position of the wire at `z=0`.
    - `mid_y`: y position of the wire at `z=0`.
    - `west_x`: x position of the west end of the wire.
    - `west_y`: y position of the west end of the wire.
    - `west_z`: z position of the west end of the wire.
    - `east_x`: x position of the east end of the wire.
    - `east_y`: y position of the east end of the wire.
    - `east_z`: z position of the east end of the wire.

    Parameters:
        gid: The gid of the wire.
        with_pos: Whether to include the position information.

    Returns:
        The parsed result.
    """
    layer = mdc_gid_to_layer(gid)
    wire = mdc_gid_to_wire(gid)

    res = {
        "gid": gid,
        "layer": layer,
        "wire": wire,
        "stereo": mdc_gid_to_stereo(gid),
        "is_stereo": mdc_gid_to_is_stereo(gid),
        "superlayer": mdc_gid_to_superlayer(gid),
    }

    if with_pos:
        west_x = mdc_gid_to_west_x(gid)
        west_y = mdc_gid_to_west_y(gid)
        east_x = mdc_gid_to_east_x(gid)
        east_y = mdc_gid_to_east_y(gid)
        res["mid_x"] = (west_x + east_x) / 2
        res["mid_y"] = (west_y + east_y) / 2
        res["west_x"] = west_x
        res["west_y"] = west_y
        res["west_z"] = mdc_gid_to_west_z(gid)
        res["east_x"] = east_x
        res["east_y"] = east_y
        res["east_z"] = mdc_gid_to_east_z(gid)

    if isinstance(gid, ak.Array):
        return ak.zip(res)
    else:
        return res


def parse_mdc_digi_id(
    mdc_digi_id: IntLike,
    with_pos: bool = False,
) -> Union[IntLike, dict[str, IntLike]]:
    """
    Parse MDC digi ID.

    When `mdc_digi_id` is an `ak.Array`, the result is an `ak.Array`, otherwise it is a `dict`.

    Keys of the output:

    - `gid`: Global ID of the wire.
    - `wire`: Local wire number.
    - `layer`: Layer number.
    - `stereo`: Stereo type. 0 for axial, -1 for `phi_west < phi_east`, 1 for `phi_west > phi_east`.
    - `is_stereo`: Whether the wire is a stereo wire.
    - `superlayer`: Superlayer number.

    Optional keys of the output when `with_pos` is `True`:

    - `mid_x`: x position of the wire at `z=0`.
    - `mid_y`: y position of the wire at `z=0`.
    - `west_x`: x position of the west end of the wire.
    - `west_y`: y position of the west end of the wire.
    - `west_z`: z position of the west end of the wire.
    - `east_x`: x position of the east end of the wire.
    - `east_y`: y position of the east end of the wire.
    - `east_z`: z position of the east end of the wire.

    Parameters:
        mdc_digi_id: The MDC digi ID.
        with_pos: Whether to include the position information.

    Returns:
        The parsed result.
    """
    wire = digi_id.mdc_id_to_wire(mdc_digi_id)
    layer = digi_id.mdc_id_to_layer(mdc_digi_id)
    gid = get_mdc_gid(layer, wire)
    return parse_mdc_gid(gid, with_pos)


def parse_mdc_digi(mdc_digi: ak.Record, with_pos: bool = False) -> ak.Record:
    """
    Parse MDC raw digi array. The raw digi array should contain [`m_intId`,
    `m_timeChannel`, `m_chargeChannel`, `m_trackIndex`, `m_overflow`] fields.

    Fields of the output:

    - `gid`: Global ID of the wire.
    - `wire`: Local wire number.
    - `layer`: Layer number.
    - `stereo`: Stereo type. 0 for axial, -1 for `phi_west < phi_east`, 1 for `phi_west > phi_east`.
    - `is_stereo`: Whether the wire is a stereo wire.
    - `superlayer`: Superlayer number.
    - `charge_channel`: Charge channel.
    - `time_channel`: Time channel.
    - `track_index`: Track index.
    - `overflow`: Overflow flag.
    - `digi_id`: Raw digi ID.

    Optional fields of the output when `with_pos` is `True`:

    - `mid_x`: x position of the wire at `z=0`.
    - `mid_y`: y position of the wire at `z=0`.
    - `west_x`: x position of the west end of the wire.
    - `west_y`: y position of the west end of the wire.
    - `west_z`: z position of the west end of the wire.
    - `east_x`: x position of the east end of the wire.
    - `east_y`: y position of the east end of the wire.
    - `east_z`: z position of the east end of the wire.

    Parameters:
        mdc_digi: The MDC raw digi array.
        with_pos: Whether to include the position information.

    Returns:
        The parsed MDC digi array.
    """
    gid = parse_mdc_digi_id(mdc_digi["m_intId"], with_pos=with_pos)

    res = {
        "gid": gid["gid"],
        "wire": gid["wire"],
        "layer": gid["layer"],
        "stereo": gid["stereo"],
        "is_stereo": gid["is_stereo"],
        "superlayer": gid["superlayer"],
        "charge_channel": mdc_digi["m_chargeChannel"],
        "time_channel": mdc_digi["m_timeChannel"],
        "track_index": mdc_digi["m_trackIndex"],
        "overflow": mdc_digi["m_overflow"],
        "digi_id": mdc_digi["m_intId"],
    }

    if with_pos:
        res["mid_x"] = gid["mid_x"]
        res["mid_y"] = gid["mid_y"]
        res["west_x"] = gid["west_x"]
        res["west_y"] = gid["west_y"]
        res["west_z"] = gid["west_z"]
        res["east_x"] = gid["east_x"]
        res["east_y"] = gid["east_y"]
        res["east_z"] = gid["east_z"]

    return ak.zip(res)


###############################################################################
#                                     TOF                                     #
###############################################################################
def parse_tof_digi_id(
    tof_digi_id: IntLike,
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse TOF digi ID.
    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:

    - `part`: The part number. `0,1,2` for scintillator endcap0, barrel, endcap1; `3,4` for MRPC endcap0, endcap1.
    - `layer_or_module`: The scintillator layer or MRPC module number, based on the part number.
    - `phi_or_strip`: The scintillator phi or MRPC strip ID, based on the part number.
    - `end`: The readout end ID.

    The return value is based on the part number.

    Rows where `part < 3` are scintillator and `layer_or_module` represents layer number, `phi_or_strip` represents phi number.

    Rows where `part >= 3` are MRPC and `layer_or_module` represents module number, `phi_or_strip` represents strip ID.

    Parameters:
        tof_digi_id: The TOF ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed TOF ID.

    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(tof_digi_id, ak.Array):
        tof_digi_id = ak.flatten(tof_digi_id)

    part = digi_id.tof_id_to_part(tof_digi_id)
    res = {
        "part": part,
        "layer_or_module": digi_id.tof_id_to_layer_or_module(tof_digi_id, part),
        "phi_or_strip": digi_id.tof_id_to_phi_or_strip(tof_digi_id, part),
        "end": digi_id.tof_id_to_end(tof_digi_id),
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res


###############################################################################
#                                     EMC                                     #
###############################################################################
def parse_emc_gid(gid: IntLike, with_pos: bool = True) -> Union[IntLike, dict[str, IntLike]]:
    """
    Parse the gid of EMC crystals. "gid" is the global ID of the crystal, ranges from 0 to 6239.
    When `gid` is an `ak.Array`, the result is an `ak.Array`, otherwise it is a `dict`.

    Keys of the output:

    - `gid`: Global ID of the crystal.
    - `part`: Part number, 0 for endcap0, 1 for barrel, 2 for endcap1.
    - `theta`: Theta number.
    - `phi`: Phi number.

    Optional keys of the output when `with_pos` is `True`:

    - `front_center_x`: x position of the front center of the crystal.
    - `front_center_y`: y position of the front center of the crystal.
    - `front_center_z`: z position of the front center of the crystal.
    - `center_x`: x position of the center of the crystal.
    - `center_y`: y position of the center of the crystal.
    - `center_z`: z position of the center of the crystal.

    !!! info
        The 8 points of the crystal will not be returned here.
        If you need the 8 points of the crystal, use `emc_gid_to_point_x`, `emc_gid_to_point_y`
        and `emc_gid_to_point_z`.

    Parameters:
        gid: The gid of the crystal.
        with_pos: Whether to include the position information.

    Returns:
        The parsed result.
    """
    part = emc_gid_to_part(gid)
    theta = emc_gid_to_theta(gid)
    phi = emc_gid_to_phi(gid)

    res = {"gid": gid, "part": part, "theta": theta, "phi": phi}

    if with_pos:
        res["front_center_x"] = emc_gid_to_front_center_x(gid)
        res["front_center_y"] = emc_gid_to_front_center_y(gid)
        res["front_center_z"] = emc_gid_to_front_center_z(gid)
        res["center_x"] = emc_gid_to_center_x(gid)
        res["center_y"] = emc_gid_to_center_y(gid)
        res["center_z"] = emc_gid_to_center_z(gid)

    if isinstance(gid, ak.Array):
        return ak.zip(res)
    else:
        return res


def parse_emc_digi_id(
    emc_digi_id: IntLike,
    with_pos: bool = False,
) -> Union[IntLike, dict[str, IntLike]]:
    """
    Parse EMC digi ID.

    When `emc_digi_id` is an `ak.Array`, the result is an `ak.Array`, otherwise it is a `dict`.

    Keys of the output:

    - `gid`: Global ID of the crystal.
    - `part`: Part number, 0 for endcap0, 1 for barrel, 2 for endcap1.
    - `theta`: Theta number.
    - `phi`: Phi number.

    Optional keys of the output when `with_pos` is `True`:

    - `front_center_x`: x position of the front center of the crystal.
    - `front_center_y`: y position of the front center of the crystal.
    - `front_center_z`: z position of the front center of the crystal.
    - `center_x`: x position of the center of the crystal.
    - `center_y`: y position of the center of the crystal.
    - `center_z`: z position of the center of the crystal.

    !!! info
        The 8 points of the crystal will not be returned here.
        If you need the 8 points of the crystal, use `emc_gid_to_point_x`, `emc_gid_to_point_y`
        and `emc_gid_to_point_z`.

    Parameters:
        emc_digi_id: The EMC digi ID.
        with_pos: Whether to include the position information.

    Returns:
        The parsed EMC digi ID.

    """
    part = digi_id.emc_id_to_module(emc_digi_id)
    theta = digi_id.emc_id_to_theta(emc_digi_id)
    phi = digi_id.emc_id_to_phi(emc_digi_id)
    gid = get_emc_gid(part, theta, phi)
    return parse_emc_gid(gid, with_pos)


def parse_emc_digi(emc_digi: ak.Record, with_pos: bool = False) -> ak.Record:
    """
    Parse EMC raw digi array. The raw digi array should contain [`m_intId`,
    `m_timeChannel`, `m_chargeChannel`, `m_trackIndex`, `m_measure`] fields.

    Fields of the output:

    - `gid`: Global ID of the crystal.
    - `part`: Part number, 0 for endcap0, 1 for barrel, 2 for endcap1.
    - `theta`: Theta number.
    - `phi`: Phi number.
    - `charge_channel`: Charge channel.
    - `time_channel`: Time channel.
    - `track_index`: Track index.
    - `measure`: Measure value.
    - `digi_id`: Raw digi ID.

    Optional fields of the output when `with_pos` is `True`:

    - `front_center_x`: x position of the front center of the crystal.
    - `front_center_y`: y position of the front center of the crystal.
    - `front_center_z`: z position of the front center of the crystal.
    - `center_x`: x position of the center of the crystal.
    - `center_y`: y position of the center of the crystal.
    - `center_z`: z position of the center of the crystal.

    Parameters:
        emc_digi: The EMC raw digi array.
        with_pos: Whether to include the position information.

    Returns:
        The parsed EMC digi array.
    """
    gid = parse_emc_digi_id(emc_digi["m_intId"], with_pos=with_pos)

    res = {
        "gid": gid["gid"],
        "part": gid["part"],
        "theta": gid["theta"],
        "phi": gid["phi"],
        "charge_channel": emc_digi["m_chargeChannel"],
        "time_channel": emc_digi["m_timeChannel"],
        "track_index": emc_digi["m_trackIndex"],
        "measure": emc_digi["m_measure"],
        "digi_id": emc_digi["m_intId"],
    }

    if with_pos:
        res["front_center_x"] = gid["front_center_x"]
        res["front_center_y"] = gid["front_center_y"]
        res["front_center_z"] = gid["front_center_z"]
        res["center_x"] = gid["center_x"]
        res["center_y"] = gid["center_y"]
        res["center_z"] = gid["center_z"]

    return ak.zip(res)


###############################################################################
#                                     MUC                                     #
###############################################################################
def parse_muc_digi_id(
    muc_digi_id: IntLike,
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse MUC digi ID.

    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:

    - `part`: The part number.
    - `segment`: The segment number.
    - `layer`: The layer number.
    - `channel`: The channel number.
    - `gap`: The gap number, which is equivalent to layer number.
    - `strip`: The strip number, which is equivalent to channel number.

    Parameters:
        muc_digi_id: The MUC digi ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed MUC digi ID.
    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(muc_digi_id, ak.Array):
        muc_digi_id = ak.flatten(muc_digi_id)

    part = digi_id.muc_id_to_part(muc_digi_id)
    segment = digi_id.muc_id_to_segment(muc_digi_id)
    layer = digi_id.muc_id_to_layer(muc_digi_id)
    channel = digi_id.muc_id_to_channel(muc_digi_id)

    res = {
        "part": part,
        "segment": segment,
        "layer": layer,
        "channel": channel,
        "gap": layer,
        "strip": channel,
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res


###############################################################################
#                                    CGEM                                     #
###############################################################################
def parse_cgem_digi_id(
    cgem_digi_id: IntLike,
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse CGEM digi ID.

    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:

    - `layer`: The layer number.
    - `sheet`: The sheet ID.
    - `strip`: The strip ID.
    - `is_x_strip`: Whether the strip is an X-strip.

    Parameters:
        cgem_digi_id: The CGEM digi ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed CGEM digi ID.
    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(cgem_digi_id, ak.Array):
        cgem_digi_id = ak.flatten(cgem_digi_id)

    res = {
        "layer": digi_id.cgem_id_to_layer(cgem_digi_id),
        "sheet": digi_id.cgem_id_to_sheet(cgem_digi_id),
        "strip": digi_id.cgem_id_to_strip(cgem_digi_id),
        "is_x_strip": digi_id.cgem_id_to_is_x_strip(cgem_digi_id),
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res


__all__ = [
    # geometry
    "get_mdc_wire_position",
    "mdc_gid_z_to_x",
    "mdc_gid_z_to_y",
    # identifier
    "get_mdc_digi_id",
    "get_tof_digi_id",
    "get_emc_digi_id",
    "get_muc_digi_id",
    "get_cgem_digi_id",
    # parser
    "parse_mdc_gid",
    "parse_mdc_digi_id",
    "parse_tof_digi_id",
    "parse_emc_digi_id",
    "parse_muc_digi_id",
    "parse_cgem_digi_id",
]
